# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Integration tests for the branch page."""

from __future__ import unicode_literals

import pytest
from kwalitee.models import CommitStatus, BranchStatus
from hamcrest import assert_that, equal_to, contains_string


@pytest.fixture(scope="function")
def branch(owner, repository, session, request):
    cs = [
        {"sha": "ef01234"},
        {"sha": "789abcd"},
        {"sha": "0123456"},
    ]
    name = "spam:wip/my-branch"
    url_template = "https://github.com/invenio/test/commits/{sha}"

    commits = []
    branches = []
    for commit in cs:
        cs = CommitStatus(repository,
                          commit["sha"],
                          url_template.format(**commit))
        commits.append(cs)
        session.add(cs)
        session.commit()

        bs = BranchStatus(commits[-1],
                          name,
                          "http://github.com/pulls/1",
                          {"commits": commits, "files": {}})
        branches.append(bs)
        session.add(bs)
        session.commit()

    def teardown():
        for bs in branches:
            session.delete(bs)
        session.commit()
        for commit in commits:
            session.delete(commit)
        session.commit()

    request.addfinalizer(teardown)
    return branches[-1]


def test_get_branch(app, owner, repository, branch):
    """GET /{account}/{repository} displays the recent commits."""

    tester = app.test_client()
    response = tester.get("/{0}/{1}/branches/{2}".format(
                          owner.name,
                          repository.name,
                          branch.name))

    assert_that(response.status_code, equal_to(200))
    body = response.get_data(as_text=True)
    assert_that(body,
                contains_string("/{0}/{1}/".format(
                                owner.name,
                                repository.name)))

    for commit in branch.content["commits"]:
        assert_that(body,
                    contains_string("/{0}/{1}/branches/{2}/{3}"
                                    .format(owner.name,
                                            repository.name,
                                            commit,
                                            branch.name)))
    assert_that(body, contains_string("Everything is OK."))


def test_get_branch_doesnt_exist(app, owner, repository):
    """GET /{account}/{repository}/branches/404 raise 404 if not found."""

    tester = app.test_client()
    response = tester.get("/{0}/{1}/branches/404".format(
                          owner.name,
                          repository.name))

    assert_that(response.status_code, equal_to(404))



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import unicode_literals

import os
import sys
import pytest
import shutil
import tempfile
import subprocess
from io import StringIO
from unittest import TestCase
from mock import patch, mock_open, MagicMock
from hamcrest import (assert_that, equal_to, has_length, has_item, has_items,
                      is_not, contains_string)

from kwalitee.hooks import (_get_component, _get_components,
                                    _get_git_author, _get_files_modified,
                                    _pre_commit, _prepare_commit_msg,
                                    post_commit_hook, pre_commit_hook)


class GetComponentTest(TestCase):
    """Testing _get_component and _get_components."""

    def test_get_component(self):
        files = {
            "invenio/base/factory.py": "base",
            "invenio/modules/oauthclient/utils.py": "oauthclient",
            "zenodo/modules/deposit/__init__.py": "deposit",
            "invenio/legacy/bibupload/engine.py": "bibupload",
            "invenio/ext/sqlalchemy/__init__.py": "sqlalchemy",
            "docs/index.rst": "docs",
            "grunt/app.js": "grunt",
            "setup.py": "global",
        }

        for filename, expected in files.items():
            assert_that(_get_component(filename), equal_to(expected), filename)

    def test_get_components(self):
        components = _get_components(("setup.py", "test.py", "grunt/app.js"))
        assert_that(components, has_length(2))
        assert_that(components, has_items("global", "grunt"))


class PreCommitTest(TestCase):
    """Testing the pre-commit actions."""

    options = {"ignore": ("D100",)}

    def test_pre_commit(self):
        errors = _pre_commit((("__init__.py", b""),
                              ("a/b/c/d/e/f/g/test.py", b""),
                              ("i/j/k/error.py", b"import os")),
                             self.options)

        assert_that(
            errors,
            has_items(
                "i/j/k/error.py: 1: L101 copyright is missing",
                "i/j/k/error.py: 1:1: F401 'os' imported but unused"))


class PrepareCommitMsgTest(TestCase):
    """Testing the preparation of commit message."""

    def _mock_open(self, data=""):
        """Creates a mock for a file with the given data and returns the mock
        and the file handler.

        NB: everytime the file is reopened, it's trucated. To be used in
        read or read then write operations.
        """
        mock = mock_open()
        filehandler = StringIO(data)

        def _open():
            if filehandler.tell() > 0:
                filehandler.truncate(0)
                filehandler.seek(0)
            return filehandler

        mock.return_value = MagicMock(spec=StringIO)
        mock.return_value.__enter__.side_effect = _open
        return mock, filehandler

    def test_prepare_commit_msg(self):
        commit_msg = "# this is a comment"
        mock, tmp_file = self._mock_open(commit_msg)
        with patch("kwalitee.hooks.open", mock, create=True):
            _prepare_commit_msg("mock", "John",
                                template="{component}: {author}")

            tmp_file.seek(0)
            new_commit_msg = "\n".join(tmp_file.readlines())

            assert_that(new_commit_msg, equal_to("unknown: John"))

    def test_prepare_commit_msg_with_one_component(self):
        commit_msg = "# this is a comment"
        mock, tmp_file = self._mock_open(commit_msg)
        with patch("kwalitee.hooks.open", mock, create=True):
            _prepare_commit_msg("mock", "John",
                                ("setup.py", "test.py"),
                                "{component}")

            tmp_file.seek(0)
            new_commit_msg = "\n".join(tmp_file.readlines())

            assert_that(new_commit_msg, equal_to("global"))

    def test_prepare_commit_msg_with_many_components(self):
        commit_msg = "# this is a comment"
        mock, tmp_file = self._mock_open(commit_msg)
        with patch("kwalitee.hooks.open", mock, create=True):
            _prepare_commit_msg("mock", "John",
                                ("setup.py",
                                 "grunt/foo.js",
                                 "docs/bar.rst"),
                                "{component}")

            tmp_file.seek(0)
            new_commit_msg = "\n".join(tmp_file.readlines())

            assert_that(new_commit_msg, contains_string("global"))
            assert_that(new_commit_msg, contains_string("grunt"))
            assert_that(new_commit_msg, contains_string("docs"))

    def test_prepare_commit_msg_aborts_if_existing(self):
        commit_msg = "Lorem ipsum"
        mock, tmp_file = self._mock_open(commit_msg)
        with patch("kwalitee.hooks.open", mock, create=True):
            _prepare_commit_msg("mock", "John")

            tmp_file.seek(0)
            new_commit_msg = "\n".join(tmp_file.readlines())

            assert_that(new_commit_msg, equal_to(commit_msg))


@pytest.fixture(scope="function")
def github(app, session, request):
    commit_msg = "dummy: test\n\n" \
                 "* foo\n" \
                 "  bar\n\n" \
                 "Signed-off-by: John Doe <john.doe@example.org>"
    cmds = (
        "git init",
        "git config user.name 'Jürg Müller'",
        "git config user.email juerg.mueller@example.org",
        "touch empty.py",
        "git add empty.py",
        "git commit -m '{0}'".format(commit_msg),
        "touch README.rst",
        "git add README.rst",
        "mkdir -p invenio/modules/testmod1/",
        "mkdir invenio/modules/testmod2/",
        "echo pass > invenio/modules/testmod1/test.py",
        "echo pass > invenio/modules/testmod2/test.py",
        "touch invenio/modules/testmod1/script.js",
        "touch invenio/modules/testmod1/style.css",
        "git add invenio/modules/testmod1/test.py",
        "git add invenio/modules/testmod2/test.py",
        "git add invenio/modules/testmod1/script.js",
        "git add invenio/modules/testmod1/style.css",
    )

    config = app.config

    app.config.update({
        "COMPONENTS": ["dummy"],
        "TRUSTED_DEVELOPERS": ["john.doe@example.org"]
    })

    path = tempfile.mkdtemp()
    cwd = os.getcwd()
    os.chdir(path)
    for command in cmds:
        proc = subprocess.Popen(command.encode("utf-8"),
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                shell=True,
                                cwd=path)
        (stdout, stderr) = proc.communicate()
        assert_that(proc.returncode, equal_to(0),
                    "{0}: {1}".format(command, stderr))

    def teardown():
        shutil.rmtree(path)
        os.chdir(cwd)
        app.config = config

    request.addfinalizer(teardown)
    return path


def test_get_files_modified(github):
    assert_that(_get_files_modified(), is_not(has_item("empty.py")))
    assert_that(_get_files_modified(),
                has_items("README.rst",
                          "invenio/modules/testmod1/test.py",
                          "invenio/modules/testmod2/test.py",
                          "invenio/modules/testmod1/script.js",
                          "invenio/modules/testmod1/style.css"))


def test_get_git_author(github):
    assert_that(_get_git_author(),
                equal_to("Jürg Müller <juerg.mueller@example.org>"))


def test_post_commit_hook(github):
    """Hook: post-commit doesn't fail"""
    stderr = sys.stderr

    sys.stderr = StringIO()
    assert_that(post_commit_hook() == 0)
    sys.stderr.seek(0)
    output = "\n".join(sys.stderr.readlines())
    sys.stderr = stderr

    assert_that(output, has_length(0))


def test_pre_commit_hook(github):
    """Hook: pre-commit fails because some copyrights are missing"""
    stderr = sys.stderr

    sys.stderr = StringIO()
    assert_that(pre_commit_hook() == 1)
    sys.stderr.seek(0)
    output = "\n".join(sys.stderr.readlines())
    sys.stderr = stderr

    assert_that(output, contains_string("kwalitee errors"))



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Integration tests for the pull_request event."""

from __future__ import absolute_import, unicode_literals

import httpretty

from flask import json
from datetime import datetime
from kwalitee.models import BranchStatus, CommitStatus
from kwalitee.tasks import pull_request
from hamcrest import (assert_that, equal_to, contains_string, has_length,
                      has_item, has_items, is_not, greater_than)

from utils import MyQueue, GPL


def test_pull_request(app, owner, repository):
    """POST /payload (pull_request) performs the checks"""
    httpretty.reset()
    queue = MyQueue()
    # Replace the default Redis queue
    app.config["queue"] = queue

    pull_request_event = {
        "action": "opened",
        "number": 1,
        "pull_request": {
            "title": "Lorem ipsum",
            "url": "https://api.github.com/pulls/1",
            "html_url": "https://github.com/pulls/1",
            "commits_url": "https://api.github.com/pulls/1/commits",
            "statuses_url": "https://api.github.com/pulls/1/statuses",
            "head": {
                "sha": "2",
                "label": "spam:wip/my-branch",
                "ref": "wip/my-branch"
            }
        },
        "repository": {
            "name": "test",
            "owner": {
                "login": "invenio"
            }
        }
    }

    commits = [
        {
            "url": "https://api.github.com/commits/1",
            "sha": "1",
            "html_url": "https://github.com/commits/1",
            "comments_url": "https://api.github.com/commits/1/comments",
            "commit": {
                "message": "herp derp"
            }
        }, {
            "url": "https://api.github.com/commits/2",
            "sha": "2",
            "html_url": "https://github.com/commits/2",
            "comments_url": "https://api.github.com/commits/2/comments",
            "commit": {
                "message": "fix all the bugs!"
            }
        }
    ]
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1/commits",
                           body=json.dumps(commits),
                           content_type="application/json")

    tester = app.test_client()
    httpretty.enable()
    response = tester.post("/payload", content_type="application/json",
                           headers=(("X-GitHub-Event", "pull_request"),
                                    ("X-GitHub-Delivery", "1")),
                           data=json.dumps(pull_request_event))
    httpretty.disable()

    assert_that(response.status_code, equal_to(200))
    body = json.loads(response.data)
    assert_that(body["payload"]["state"], equal_to("pending"))

    (fn, bs_id, pull_request_url, status_url, config) = queue.dequeue()
    assert_that(fn, equal_to(pull_request))
    assert_that(bs_id, greater_than(0))
    assert_that(pull_request_url,
                equal_to("https://api.github.com/pulls/1"))

    cs = CommitStatus.query.filter_by(repository_id=repository.id,
                                      sha="1").first()
    assert_that(cs)
    assert_that(cs.state, equal_to("pending"))

    cs = CommitStatus.query.filter_by(repository_id=repository.id,
                                      sha="2").first()
    assert_that(cs)
    assert_that(cs.state, equal_to("pending"))

    bs = BranchStatus.query.filter_by(commit_id=cs.id,
                                      name="spam:wip/my-branch").first()
    assert_that(bs)
    assert_that(bs.is_pending())
    assert_that(bs.state, equal_to("pending"))


def test_invalid_pull_request_action(app, owner, repository):
    """POST /payload (pull_request) performs the action checks"""
    httpretty.reset()
    queue = MyQueue()
    # Replace the default Redis queue
    app.config["queue"] = queue

    pull_request_event = {
        "action": "labeled",
        "number": 1,
        "pull_request": {
            "title": "Lorem ipsum",
            "url": "https://api.github.com/pulls/1",
            "html_url": "https://github.com/pulls/1",
            "commits_url": "https://api.github.com/pulls/1/commits",
            "statuses_url": "https://api.github.com/pulls/1/statuses",
            "head": {
                "sha": "2",
                "label": "spam:wip/my-branch",
                "ref": "wip/my-branch"
            }
        },
        "repository": {
            "name": "test",
            "owner": {
                "login": "invenio"
            }
        }
    }
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1/commits",
                           body=json.dumps([]),
                           content_type="application/json")

    tester = app.test_client()
    httpretty.enable()

    response = tester.post("/payload", content_type="application/json",
                           headers=(("X-GitHub-Event", "pull_request"),
                                    ("X-GitHub-Delivery", "1")),
                           data=json.dumps(pull_request_event))
    httpretty.disable()

    assert_that(response.status_code, equal_to(500))


def test_pull_request_task(app, owner, repository, session):
    """Task pull_request /pulls/1"""
    httpretty.reset()
    pull = {
        "title": "Lorem ipsum",
        "url": "https://api.github.com/pulls/1",
        "html_url": "https://github.com/pulls/1",
        "commits_url": "https://api.github.com/pulls/1/commits",
        "statuses_url": "https://api.github.com/statuses/2",
        "review_comments_url": "https://api.github.com/pulls/1/comments",
        "issue_url": "https://api.github.com/issues/1",
        "head": {
            "sha": "2",
            "label": "test:my-branch"
        },
        "base": {
            "repo": {
                "full_name": "kwalitee/test"
            },
            "ref": "testref"
        },
    }
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1",
                           body=json.dumps(pull),
                           content_type="application/json")

    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/repos/kwalitee/test/"
                           "contents/.kwalitee.yml?ref=testref",
                           status=404)

    issue = {
        "url": "https://api.github.com/issues/1",
        "html_url": "https://github.com/issues/1",
        "labels_url": "https://api.github.com/issues/1/labels{/name}",
        "id": "42",
        "number": "1",
        "labels": [{"name": "foo"},
                   {"name": "in_work"}],
        "state": "open"
    }
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/issues/1",
                           body=json.dumps(issue),
                           content_type="application/json")
    labels = [{
        "url": "https://github.com/labels/foo",
        "name": "foo",
        "color": "000000"
    }, {
        "url": "https://github.com/labels/in_review",
        "name": "in_review",
        "color": "ff0000"
    }]
    httpretty.register_uri(httpretty.PUT,
                           "https://api.github.com/issues/1/labels",
                           status=200,
                           body=json.dumps(labels),
                           content_type="application/json")
    commits = [
        {
            "url": "https://api.github.com/commits/1",
            "sha": "1",
            "html_url": "https://github.com/commits/1",
            "comments_url": "https://api.github.com/commits/1/comments",
            "commit": {
                "message": "fix all the bugs!"
            }
        }, {

            "url": "https://api.github.com/commits/2",
            "sha": "2",
            "html_url": "https://github.com/commits/2",
            "comments_url": "https://api.github.com/commits/1/comments",
            "commit": {
                "message": "herp derp"
            }
        }
    ]
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1/commits",
                           body=json.dumps(commits),
                           content_type="application/json")
    files = [{
        "filename": "spam/eggs.py",
        "status": "added",
        "raw_url": "https://github.com/raw/2/spam/eggs.py",
        "contents_url": "https://api.github.com/spam/eggs.py?ref=2"
    }, {
        "filename": "spam/herp.html",
        "status": "added",
        "raw_url": "https://github.com/raw/2/spam/herp.html",
        "contents_url": "https://api.github.com/spam/herp.html?ref=2"
    }]
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1/files",
                           body=json.dumps(files),
                           content_type="application/json")
    eggs_py = "if foo == bar:\n  print('derp')\n"
    httpretty.register_uri(httpretty.GET,
                           "https://github.com/raw/2/spam/eggs.py",
                           body=eggs_py,
                           content_type="text/plain")
    herp_html = "<!DOCTYPE html><html><title>Hello!</title></html>"
    httpretty.register_uri(httpretty.GET,
                           "https://github.com/raw/2/spam/herp.html",
                           body=herp_html,
                           content_type="text/html")
    httpretty.register_uri(httpretty.POST,
                           "https://api.github.com/commits/1/comments",
                           status=201,
                           body=json.dumps({"id": 1}),
                           content_type="application/json")
    httpretty.register_uri(httpretty.POST,
                           "https://api.github.com/commits/2/comments",
                           status=201,
                           body=json.dumps({"id": 2}),
                           content_type="application/json")
    httpretty.register_uri(httpretty.POST,
                           "https://api.github.com/pulls/1/comments",
                           status=201,
                           body=json.dumps({"id": 3}),
                           content_type="application/json")
    status = {"id": 1, "state": "success"}
    httpretty.register_uri(httpretty.POST,
                           "https://api.github.com/statuses/2",
                           status=201,
                           body=json.dumps(status),
                           content_type="application/json")

    css = []
    for commit in commits:
        css.append(CommitStatus.find_or_create(repository,
                                               commit["sha"],
                                               commit["url"]))

    bs = BranchStatus(css[-1],
                      "test:my-branch",
                      "https://github.com/pulls/1",
                      {"commits": css, "files": None})
    session.add(bs)
    session.commit()

    assert_that(css[0].is_pending())
    assert_that(css[1].is_pending())
    assert_that(bs.is_pending())

    httpretty.enable()
    pull_request(bs.id,
                 "https://api.github.com/pulls/1",
                 "http://kwalitee.invenio-software.org/status/2",
                 {"repository": repository.id})
    httpretty.disable()

    latest_requests = httpretty.HTTPretty.latest_requests
    # 7x GET pull, .kwalitee.yml, issue, commits, files, spam/eggs.py,
    # spam/herp.html
    # 5x POST comments (1 message + 2 files), status
    # 1x PUT labels
    assert_that(len(latest_requests), equal_to(12), "7x GET + 4x POST + 1 PUT")

    expected_requests = [
        "",
        "",
        "",
        "missing component name",  # "signature is missing",
        "",
        "",
        "",
        "F821 undefined name",
        "L101 copyright is missing",
        "/status/2",
        "",
        "in_review"
    ]
    for expected, request in zip(expected_requests, latest_requests):
        assert_that(str(request.body), contains_string(expected))

    body = json.loads(latest_requests[-3].body)
    assert_that(latest_requests[-3].headers["authorization"],
                equal_to("token {0}".format(owner.token)))
    assert_that(body["state"], equal_to("error"))

    cs = CommitStatus.query.filter_by(repository_id=repository.id).all()

    assert_that(cs, has_length(2))
    assert_that(cs[0].content["message"],
                has_item("1: M110 missing component name"))
    assert_that(cs[1].content["message"],
                has_item("1: M100 needs more reviewers"))

    bs = BranchStatus.query.filter_by(commit_id=cs[1].id,
                                      name="test:my-branch").first()

    assert_that(bs)
    assert_that(bs.content["commits"], has_items("1", "2"))
    assert_that(bs.errors, equal_to(12))
    assert_that(
        bs.content["files"]["spam/eggs.py"]["errors"],
        has_item("2:3: E111 indentation is not a multiple of four"))


def test_wip_pull_request_task(app, owner, repository, session):
    """Task pull_request /pulls/1 is work in progress"""
    httpretty.reset()
    pull = {
        "title": "WIP Lorem ipsum",
        "url": "https://api.github.com/pulls/1",
        "html_url": "https://github.com/pulls/1",
        "commits_url": "https://api.github.com/pulls/1/commits",
        "statuses_url": "https://api.github.com/statuses/1",
        "review_comments_url": "https://api.github.com/pulls/1/comments",
        "issue_url": "https://api.github.com/issues/1",
        "head": {
            "sha": "2",
            "label": "test:my-branch"
        },
        "base": {
            "repo": {
                "full_name": "kwalitee/test"
            },
            "ref": "testref"
        },
    }
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1",
                           body=json.dumps(pull),
                           content_type="application/json")

    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/repos/kwalitee/test/"
                           "contents/.kwalitee.yml?ref=testref",
                           status=404)

    commits = [
        {
            "url": "https://api.github.com/commits/1",
            "sha": "1",
            "html_url": "https://github.com/commits/1",
            "comments_url": "https://api.github.com/commits/1/comments",
            "commit": {
                "message": "fix all the bugs!"
            }
        }, {

            "url": "https://api.github.com/commits/2",
            "sha": "2",
            "html_url": "https://github.com/commits/2",
            "comments_url": "https://api.github.com/commits/2/comments",
            "commit": {
                "message": "herp derp"
            }
        }
    ]
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1/commits",
                           body=json.dumps(commits),
                           content_type="application/json")
    issue = {
        "url": "https://api.github.com/issues/1",
        "html_url": "https://github.com/issues/1",
        "labels_url": "https://api.github.com/issues/1/labels{/name}",
        "id": "42",
        "number": "1",
        "labels": [{"name": "foo"},
                   {"name": "in_integration"}],
        "state": "open"
    }
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/issues/1",
                           body=json.dumps(issue),
                           content_type="application/json")
    labels = [{
        "url": "https://github.com/labels/foo",
        "name": "foo",
        "color": "000000"
    }, {
        "url": "https://github.com/labels/in_review",
        "name": "in_review",
        "color": "ff0000"
    }]
    httpretty.register_uri(httpretty.PUT,
                           "https://api.github.com/issues/1/labels",
                           status=200,
                           body=json.dumps(issue),
                           content_type="application/json")

    css = []
    for commit in commits:
        css.append(CommitStatus.find_or_create(repository,
                                               commit["sha"],
                                               commit["url"]))
    bs = BranchStatus(css[-1],
                      "test:my-branch",
                      "https://github.com/pulls/1",
                      {"commits": css, "files": None})
    session.add(bs)
    session.commit()
    assert_that(bs.is_pending())

    httpretty.enable()
    pull_request(bs.id,
                 "https://api.github.com/pulls/1",
                 "http://kwalitee.invenio-software.org/status/1",
                 {"repository": repository.id})
    httpretty.disable()

    latest_requests = httpretty.HTTPretty.latest_requests
    # 4x GET pull, .kwalitee.yml, commits, issue
    # 1x POST labels
    assert_that(len(latest_requests), equal_to(5), "4x GET + 1x POST")

    expected_requests = [
        "",
        "",
        "",
        "",
        "in_work"
    ]
    for expected, request in zip(expected_requests, latest_requests):
        assert_that(str(request.body), contains_string(expected))

    labels = json.loads(latest_requests[-1].body)
    assert_that(labels, has_items("in_work", "foo"))
    assert_that(labels, is_not(has_item("in_review")))


def test_pep8_pull_request_task(app, owner, repository, session):
    """Task pull_request /pulls/1 with pep8 errors"""
    httpretty.reset()
    pull = {
        "title": "Lorem ipsum",
        "url": "https://api.github.com/pulls/1",
        "html_url": "https://github.com/pulls/1",
        "commits_url": "https://api.github.com/pulls/1/commits",
        "statuses_url": "https://api.github.com/statuses/1",
        "issue_url": "https://api.github.com/issues/1",
        "review_comments_url": "https://api.github.com/pulls/1/comments",
        "head": {
            "sha": "1",
            "label": "test:my-branch"
        },
        "base": {
            "repo": {
                "full_name": "kwalitee/test"
            },
            "ref": "testref"
        },
    }
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1",
                           body=json.dumps(pull),
                           content_type="application/json")

    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/repos/kwalitee/test/"
                           "contents/.kwalitee.yml?ref=testref",
                           status=404)

    issue = {
        "url": "https://api.github.com/issues/1",
        "html_url": "https://github.com/issues/1",
        "labels_url": "https://api.github.com/issues/1/labels{/name}",
        "id": "42",
        "number": "1",
        "labels": [{"name": "foo"},
                   {"name": "in_work"}],
        "state": "open"
    }
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/issues/1",
                           body=json.dumps(issue),
                           content_type="application/json")
    labels = [{
        "url": "https://github.com/labels/foo",
        "name": "foo",
        "color": "000000"
    }, {
        "url": "https://github.com/labels/in_review",
        "name": "in_review",
        "color": "ff0000"
    }]
    httpretty.register_uri(httpretty.PUT,
                           "https://api.github.com/issues/1/labels",
                           status=200,
                           body=json.dumps(labels),
                           content_type="application/json")
    commits = [
        {
            "url": "https://api.github.com/commits/2",
            "sha": "2",
            "html_url": "https://github.com/commits/2",
            "comments_url": "https://api.github.com/commits/1/comments",
            "commit": {
                "message": "herp: derp\r\n\r\nSigned-off-by: John Doe "
                           "<john.doe@example.org>"
            }
        }
    ]
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1/commits",
                           body=json.dumps(commits),
                           content_type="application/json")
    files = [{
        "filename": "spam/eggs.py",
        "status": "added",
        "raw_url": "https://github.com/raw/2/spam/eggs.py",
        "contents_url": "https://api.github.com/spam/eggs.py?ref=1"
    }]
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1/files",
                           body=json.dumps(files),
                           content_type="application/json")
    eggs_py = "if foo == bar:\n  print('derp')\n"
    httpretty.register_uri(httpretty.GET,
                           "https://github.com/raw/2/spam/eggs.py",
                           body=eggs_py,
                           content_type="text/plain")
    httpretty.register_uri(httpretty.POST,
                           "https://api.github.com/commits/1/comments",
                           status=201,
                           body=json.dumps({"id": 1}),
                           content_type="application/json")
    httpretty.register_uri(httpretty.POST,
                           "https://api.github.com/pulls/1/comments",
                           status=201,
                           body=json.dumps({"id": 3}),
                           content_type="application/json")
    status = {"id": 1, "state": "success"}
    httpretty.register_uri(httpretty.POST,
                           "https://api.github.com/statuses/1",
                           status=201,
                           body=json.dumps(status),
                           content_type="application/json")

    cs = []
    for commit in commits:
        cs.append(CommitStatus.find_or_create(repository,
                                              commit["sha"],
                                              commit["url"]))

    bs = BranchStatus(cs[0],
                      "test:my-branch",
                      "https://github.com/pulls/1",
                      {"commits": cs, "files": None})
    session.add(bs)
    session.commit()
    assert_that(bs.is_pending())

    httpretty.enable()
    pull_request(bs.id,
                 "https://api.github.com/pulls/1",
                 "http://kwalitee.invenio-software.org/status/1",
                 {"TRUSTED_DEVELOPERS": ["john.doe@example.org"],
                  "COMPONENTS": ["herp"],
                  "SIGNATURES": ["Signed-off-by"]})
    httpretty.disable()

    latest_requests = httpretty.HTTPretty.latest_requests
    # 6x GET pull, .kwalitee.yml, issue, commits, 1 file, spam/eggs.py
    # 3x POST comments (1xfile), status, label
    assert_that(len(latest_requests), equal_to(9), "6x GET + 3x POST")

    expected_requests = [
        "",
        "",
        "",
        "",
        "",
        "D100 Missing docstring in public module",
        "/status/1",
        "",
        "in_review",
    ]
    for expected, request in zip(expected_requests, latest_requests):
        assert_that(str(request.body), contains_string(expected))

    body = json.loads(latest_requests[-3].body)
    assert_that(latest_requests[-3].headers["authorization"],
                equal_to("token {0}".format(owner.token)))
    assert_that(body["state"], equal_to("error"))

    cs = CommitStatus.query.filter_by(repository_id=repository.id,
                                      ).all()
    assert_that(cs, has_length(1))
    assert_that(cs[0].content["files"] is None)
    assert_that(cs[0].sha, equal_to("2"))
    assert_that(cs[0].is_pending(), equal_to(False))

    bs = BranchStatus.query.filter_by(commit_id=cs[0].id,
                                      name="test:my-branch").first()

    assert_that(bs)
    assert_that(bs.errors, equal_to(5))
    assert_that(
        bs.content["files"]["spam/eggs.py"]["errors"],
        has_item("2:3: E111 indentation is not a multiple of four"))


def test_okay_pull_request_task(app, owner, repository, session):
    """Task pull_request /pulls/1 with pep8 errors"""
    httpretty.reset()
    pull = {
        "title": "Lorem ipsum",
        "url": "https://api.github.com/pulls/1",
        "html_url": "https://github.com/pulls/1",
        "commits_url": "https://api.github.com/pulls/1/commits",
        "statuses_url": "https://api.github.com/statuses/1",
        "issue_url": "https://api.github.com/issues/1",
        "review_comments_url": "https://api.github.com/pulls/1/comments",
        "head": {
            "sha": "1",
            "label": "test:my-branch"
        },
        "base": {
            "repo": {
                "full_name": "kwalitee/test"
            },
            "ref": "testref"
        },
    }
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1",
                           body=json.dumps(pull),
                           content_type="application/json")

    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/repos/kwalitee/test/"
                           "contents/.kwalitee.yml?ref=testref",
                           status=404)

    issue = {
        "url": "https://apigithub.com/issues/1",
        "html_url": "https://github.com/issues/1",
        "labels_url": "https://api.github.com/issues/1/labels{/name}",
        "id": "42",
        "number": "1",
        "labels": [{"name": "foo"},
                   {"name": "in_work"}],
        "state": "open"
    }
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/issues/1",
                           body=json.dumps(issue),
                           content_type="application/json")
    labels = [{
        "url": "https://github.com/labels/foo",
        "name": "foo",
        "color": "000000"
    }, {
        "url": "https://github.com/labels/in_review",
        "name": "in_review",
        "color": "ff0000"
    }]
    httpretty.register_uri(httpretty.PUT,
                           "https://api.github.com/issues/1/labels",
                           status=200,
                           body=json.dumps(labels),
                           content_type="application/json")
    commits = [
        {
            "url": "https://api.github.com/commits/1",
            "sha": "1",
            "html_url": "https://github.com/commits/1",
            "comments_url": "https://api.github.com/commits/1/comments",
            "commit": {
                "message": "herp: derp\r\n\r\nSigned-off-by: John Doe "
                           "<john.doe@example.org>"
            }
        }
    ]
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1/commits",
                           body=json.dumps(commits),
                           content_type="application/json")
    files = [{
        "filename": "eggs/__init__.py",
        "status": "added",
        "raw_url": "https://github.com/raw/1/eggs/__init__.py",
        "contents_url": "https://api.github.com/eggs/__init__.py?ref=1"
    }]
    httpretty.register_uri(httpretty.GET,
                           "https://api.github.com/pulls/1/files",
                           body=json.dumps(files),
                           content_type="application/json")
    init_py = GPL.format(datetime.now().year, '#')
    httpretty.register_uri(httpretty.GET,
                           "https://github.com/raw/1/eggs/__init__.py",
                           body=init_py,
                           content_type="text/plain")
    httpretty.register_uri(httpretty.POST,
                           "https://api.github.com/commits/1/comments",
                           status=201,
                           body=json.dumps({"id": 1}),
                           content_type="application/json")
    httpretty.register_uri(httpretty.POST,
                           "https://api.github.com/pulls/1/comments",
                           status=201,
                           body=json.dumps({"id": 3}),
                           content_type="application/json")
    status = {"id": 1, "state": "success"}
    httpretty.register_uri(httpretty.POST,
                           "https://api.github.com/statuses/1",
                           status=201,
                           body=json.dumps(status),
                           content_type="application/json")

    cs = []
    for commit in commits:
        cs.append(CommitStatus.find_or_create(repository,
                                              commit["sha"],
                                              commit["url"]))
    bs = BranchStatus(cs[-1],
                      "test:my-branch",
                      "http://github.com/pulls/1",
                      {"commits": cs, "files": None})
    session.add(bs)
    session.commit()

    httpretty.enable()
    pull_request(bs.id,
                 "https://api.github.com/pulls/1",
                 "http://kwalitee.invenio-software.org/status/1",
                 {"TRUSTED_DEVELOPERS": ["john.doe@example.org"],
                  "COMPONENTS": ["herp"],
                  "SIGNATURES": ["Signed-off-by"],
                  "IGNORE": ["E265", "D100"]})
    httpretty.disable()

    latest_requests = httpretty.HTTPretty.latest_requests
    # 5x GET pull, .kwalitee.yml, issue, commits, 1 file, spam/eggs.py
    # 2x POST status, label
    assert_that(len(latest_requests), equal_to(8), "6x GET + 2x POST")

    expected_requests = [
        "",
        "",
        "",
        "",
        "",
        "/status/1",
        "",
        "in_integration",
    ]
    for expected, request in zip(expected_requests, latest_requests):
        assert_that(str(request.body), contains_string(expected))

    body = json.loads(latest_requests[-3].body)
    assert_that(latest_requests[-3].headers["authorization"],
                equal_to("token {0}".format(owner.token)))
    assert_that(body["state"], equal_to("success"))

    cs = CommitStatus.query.filter_by(repository_id=repository.id).all()
    assert_that(cs, has_length(1))
    assert_that(cs[0].content["files"] is None)
    assert_that(cs[0].is_pending(), equal_to(False))

    bs = BranchStatus.query.filter_by(commit_id=cs[0].id,
                                      name="test:my-branch").first()

    assert_that(bs.errors, equal_to(0))
    assert_that(bs.content["commits"], has_length(1))
    assert_that(bs.content["files"]["eggs/__init__.py"]["errors"],
                has_length(0))


def test_known_pull_request_task(app, owner, repository, session):
    """Task pull_request /pulls/1 that already exists."""
    httpretty.reset()
    cs1 = CommitStatus(repository,
                       "1",
                       "https://github.com/pulls/1",
                       {"message": [], "files": {}})
    cs2 = CommitStatus(repository,
                       "2",
                       "https://github.com/pulls/2",
                       {"message": [], "files": {}})
    session.add(cs1)
    session.add(cs2)
    session.commit()

    bs = BranchStatus(cs2,
                      "test:my-branch",
                      "https://github.com/pulls/1",
                      {"commits": [cs1, cs2], "files": {}})
    session.add(bs)
    session.commit()
    assert_that(bs.is_pending(), equal_to(False))

    httpretty.enable()
    pull_request(bs.id,
                 "https://api.github.com/pulls/1",
                 "http://kwalitee.invenio-software.org/status/2",
                 {"ACCESS_TOKEN": "deadbeef"})
    httpretty.disable()

    latest_requests = httpretty.HTTPretty.latest_requests
    assert_that(len(latest_requests), equal_to(0),
                "No requests are expected")



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""kwalitee factory for Flask application."""


from __future__ import unicode_literals

import os

from flask import Flask, Config
from flask_sqlalchemy import SQLAlchemy
from werkzeug.routing import BaseConverter

class ShaConverter(BaseConverter):

    """Werkzeug routing converter for sha-1 (truncated or full)."""

    regex = r'(?!/)(?:[a-fA-F0-9]{40}|[a-fA-F0-9]{7})'
    weight = 150


def create_app(name=None, config=None):
    """Create the Flask application."""
    name = 'kwalitee' if name is None else name
    app = Flask(name, template_folder="templates", static_folder="static",
                instance_relative_config=True,
                instance_path=os.environ.get("KWALITEE_INSTANCE_PATH"))

    # Load default configuration
    app.config.from_object("kwalitee.config")

    # Load kwalitee.cfg from instance folder
    app.config.from_pyfile("kwalitee.cfg", silent=True)
    app.config.from_envvar("KWALITEE_CONFIG", silent=True)
    app.config.update(config or {})

    # Create instance path
    if not os.path.exists(app.instance_path):
        os.makedirs(app.instance_path)

    # Setting up the database
    from .import models
    models.init_app(app)

    # Setting up RQ
    from . import worker
    worker.init_app(app)

    # Setting up the views
    app.url_map.converters['sha'] = ShaConverter
    from . import views
    app.add_url_rule("/", "index", views.index, methods=["GET"])
    app.add_url_rule("/<account>/", "account", views.account, methods=["GET"])
    app.add_url_rule("/<account>/<repository>/", "repository", views.repository,
                     methods=["GET"])
    app.add_url_rule("/<account>/<repository>/commits/<sha:sha>/", "commit",
                     views.commit, methods=["GET"])
    app.add_url_rule("/<account>/<repository>/branches/<sha:sha>/<path:branch>",
                     "branch_status", views.branch_status, methods=["GET"])
    app.add_url_rule("/<account>/<repository>/branches/<path:branch>", "branch",
                     views.branch, methods=["GET"])
    app.add_url_rule("/payload", "payload", views.payload, methods=["POST"])
    # legacy
    app.add_url_rule("/status/<sha>", "status", views.status, methods=["GET"])

    if app.debug:
        from werkzeug.debug import DebuggedApplication
        app.wsgi_app = DebuggedApplication(app.wsgi_app, evalex=True)

    return app



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Database models to persist the data over time."""

from __future__ import unicode_literals

import os

from datetime import datetime
from flask import json
from flask_sqlalchemy import SQLAlchemy


# Storing states as integers so string can be changed/l10n later
STATE_PENDING = 0
STATE_SUCCESS = 1
STATE_ERROR = 2

STATES = {
    STATE_PENDING: "pending",
    STATE_SUCCESS: "success",
    STATE_ERROR: "error"
}


db = SQLAlchemy()


def init_app(app):
    """Initialize the Flask app with db."""
    database = os.path.join(app.instance_path,
                            app.config.get("DATABASE_NAME", "database"))
    database = "{0}.db".format(database)
    app.config.setdefault("DATABASE", database)
    app.config.setdefault("SQLALCHEMY_DATABASE_URI",
                          os.environ.get("SQLALCHEMY_DATABASE_URI",
                                         "sqlite:///{0}".format(database)))
    db.app = app
    db.init_app(app)

    if not db.engine.has_table(Account.__tablename__):
        db.create_all()
    return db


class Account(db.Model):

    """Github account."""

    id = db.Column(db.Integer(), primary_key=True, autoincrement=True)
    """Identifier"""
    name = db.Column(db.UnicodeText(), unique=True, nullable=False)
    """Username."""
    email = db.Column(db.UnicodeText())
    """Email."""
    token = db.Column(db.UnicodeText())
    """API Token."""
    created_at = db.Column(db.DateTime(), nullable=False, default=datetime.now)
    """Date of creation."""
    updated_at = db.Column(db.DateTime(), nullable=False, default=datetime.now,
                           onupdate=datetime.now)
    """Date of last modification."""

    def __init__(self, name, email=None, token=None):
        """Initialize the account."""
        self.name = name
        self.email = email
        self.token = token

    def __repr__(self):
        """String representation of the account."""
        return "<Account ({0.id}, {0.name}, {0.email})>".format(self)

    @classmethod
    def create(cls, name, email=None, token=None):
        """Create and commit and new account."""
        acc = cls(name, email, token)
        db.session.add(acc)
        db.session.commit()
        return acc

    @classmethod
    def find_or_create(cls, name, email=None, token=None):
        """Find or create an account."""
        acc = cls.query.filter_by(name=name).first()
        if not acc:
            return cls.create(name, email, token)
        return acc

    @classmethod
    def update_or_create(cls, name, email=None, token=None):
        """Modify or create an account."""
        acc = cls.query.filter_by(name=name).first()
        if not acc:
            return cls.create(name, email, token)
        else:
            if email and acc.email != email:
                acc.email = email
            if token and acc.token != token:
                acc.token = token
            if db.session.is_modified(acc):
                db.session.add(acc)
                db.session.commit()
            return acc


class Repository(db.Model):

    """Github repository."""

    __table_args__ = db.UniqueConstraint('owner_id', 'name'),

    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    owner_id = db.Column(db.Integer(),
                         db.ForeignKey("account.id"),
                         nullable=False)
    name = db.Column(db.UnicodeText(), nullable=False)
    created_at = db.Column(db.DateTime(), nullable=False, default=datetime.now)
    updated_at = db.Column(db.DateTime(), nullable=False, default=datetime.now,
                           onupdate=datetime.now)

    owner = db.relationship("Account",
                            backref=db.backref("repositories",
                                               cascade="all",
                                               order_by=db.asc(name),
                                               lazy="dynamic"))

    @property
    def fullname(self):
        """Get the fullname of the repository."""
        return "{0}/{1}".format(self.owner.name, self.name)

    def __init__(self, owner, name):
        """Initialize the repository."""
        self.owner = owner
        self.name = name

    def __repr__(self):
        """String representation of the repository."""
        return "<Repository ({0}, {1})>".format(self.id, self.fullname)

    @classmethod
    def find_or_create(cls, owner, name):
        """Find or create a repository."""
        repo = cls.query.filter_by(name=name, owner_id=owner.id).first()
        if not repo:
            repo = cls(owner, name)
            db.session.add(repo)
            db.session.commit()
        return repo


class CommitStatus(db.Model):

    """Status of a push."""

    __table_args__ = db.UniqueConstraint('repository_id', 'sha'),

    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    repository_id = db.Column(db.Integer(),
                              db.ForeignKey("repository.id"),
                              nullable=False)
    sha = db.Column(db.UnicodeText(), nullable=False)
    url = db.Column(db.UnicodeText(), nullable=False)
    _state = db.Column(db.Integer(), nullable=False)
    _errors = db.Column(db.Integer(), nullable=False)
    _content = db.Column(db.UnicodeText())
    created_at = db.Column(db.DateTime(), nullable=False, default=datetime.now)
    updated_at = db.Column(db.DateTime(), nullable=False, default=datetime.now,
                           onupdate=datetime.now)

    repository = db.relationship("Repository",
                                 backref=db.backref("commit_statuses",
                                                    cascade="all",
                                                    order_by=db.desc(id),
                                                    lazy="dynamic"))

    @property
    def errors(self):
        """Get the number of errors found."""
        return self._errors

    @property
    def state(self):
        """Get the state."""
        return STATES[self._state]

    def get_content(self):
        """Get the content of the status."""
        return json.loads(self._content)

    def set_content(self, value):
        """Set the content of the status."""
        self._errors = len(value["message"])
        if value["files"] is not None:
            for ferrors in value["files"].values():
                self._errors += len(ferrors["errors"] or [])

        if self._errors:
            self._state = STATE_ERROR
        else:
            self._state = STATE_SUCCESS

        self._content = json.dumps(value, ensure_ascii=False)
    content = property(get_content, set_content)

    def __init__(self, repository, sha, url, content=None):
        """Initialize the commit status."""
        self.repository = repository
        self.sha = sha
        self.url = url
        self._errors = 0
        if content:
            self.content = content
        else:
            self.content = {"message": [], "files": None}
            self._state = STATE_PENDING

    def __repr__(self):
        """String representation of the commit status."""
        return "<CommitStatus ({0.id}, {0.repository_id}, {0.sha}, " \
               "{0.state})>" \
               .format(self)

    def is_pending(self):
        """Return True is the commit status hasn't been checked yet."""
        return self._state == STATE_PENDING

    @classmethod
    def find_or_create(cls, repository, sha, url):
        """Find or create a commit status."""
        cs = cls.query.filter_by(repository_id=repository.id, sha=sha).first()
        if not cs:
            cs = CommitStatus(repository, sha, url)
            db.session.add(cs)
            db.session.commit()

        return cs


class BranchStatus(db.Model):

    """Status of a pull request."""

    __table_args__ = db.UniqueConstraint('commit_id', 'name'),

    id = db.Column(db.Integer(), primary_key=True, autoincrement=True)
    commit_id = db.Column(db.Integer(),
                          db.ForeignKey("commit_status.id"),
                          nullable=False)
    name = db.Column(db.UnicodeText(), nullable=False)
    url = db.Column(db.UnicodeText(), nullable=False)
    _state = db.Column(db.Integer(), nullable=False)
    _errors = db.Column(db.Integer(), nullable=False)
    _content = db.Column(db.UnicodeText())
    created_at = db.Column(db.DateTime(), nullable=False, default=datetime.now)
    updated_at = db.Column(db.DateTime(), nullable=False, default=datetime.now,
                           onupdate=datetime.now)

    commit = db.relationship("CommitStatus",
                             backref=db.backref("branch_statuses",
                                                cascade="all",
                                                order_by=db.desc(id),
                                                lazy="dynamic"))

    @property
    def errors(self):
        """Get the number of errors found."""
        return self._errors

    @property
    def state(self):
        """Get the state."""
        return STATES[self._state]

    def get_content(self):
        """Get the content of the status."""
        return json.loads(self._content)

    def set_content(self, value):
        """Set the content of the status."""
        c = {"commits": [],
             "files": value.get("files", {})}
        if "files" in value and value["files"]:
            for ferrors in value["files"].values():
                self._errors += len(ferrors["errors"] or [])
        for commit in value["commits"]:
            if not isinstance(commit, CommitStatus):
                # FIXME potentially unnecessary heavy operation
                commit = CommitStatus.query.filter_by(
                    repository_id=self.commit.repository_id,
                    sha=commit
                ).first()
            self._errors += len(commit.content["message"])
            c["commits"].append(commit.sha)

        if "files" not in value or value["files"] is None:
            self._state = STATE_PENDING
        elif self._errors:
            self._state = STATE_ERROR
        else:
            self._state = STATE_SUCCESS

        self._content = json.dumps(c, ensure_ascii=False)
    content = property(get_content, set_content)

    def __init__(self, commit, name, url, content=None):
        """Initialize a branch status."""
        self.commit = commit
        self.name = name
        self.url = url
        self._errors = 0
        if content:
            self.content = content
        else:
            self.content = {"commits": [], "files": {}}
            self._state = STATE_PENDING

    def __repr__(self):
        """String representation of a branch status."""
        return "<BranchStatus ({0.id}, {0.commit_id}, {0.name}, {0.state})>" \
               .format(self)

    def is_pending(self):
        """Return True is the commit status hasn't been checked yet."""
        return self._state == STATE_PENDING

    @classmethod
    def find_or_create(cls, commit, name, url, content=None):
        """Find or create a commit status."""
        bs = cls.query.filter_by(commit_id=commit.id, name=name).first()
        if not bs:
            bs = BranchStatus(commit, name, url, content)
            db.session.add(bs)
            db.session.commit()

        return bs



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Mocks and mixins for the tests."""

from __future__ import unicode_literals


GPL = """
{1} This file is part of kwalitee
{1} Copyright (C) {0} CERN.
{1}
{1} kwalitee is free software; you can redistribute it and/or
{1} modify it under the terms of the GNU General Public License as
{1} published by the Free Software Foundation; either version 2 of the
{1} License, or (at your option) any later version.
{1}
{1} kwalitee is distributed in the hope that it will be useful, but
{1} WITHOUT ANY WARRANTY; without even the implied warranty of
{1} MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
{1} General Public License for more details.
{1}
{1} You should have received a copy of the GNU General Public License
{1} along with kwalitee; if not, write to the Free Software Foundation,
{1} Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
{1}
{1} In applying this licence, CERN does not waive the privileges and immunities
{1} granted to it by virtue of its status as an Intergovernmental Organization
{1} or submit itself to any jurisdiction.
"""


class MyQueue(object):

    """Queue mock to use in place of the RQ queue.

    .. seealso:: `RQ <http://python-rq.org/docs/>`_
    """

    def __init__(self):
        """Initialize  an empty queue."""
        self.queue = []

    def __len__(self):
        """Length of the queue."""
        return len(self.queue)

    def dequeue(self):
        """Remove one item from the queue."""
        return self.queue.pop()

    def enqueue(self, *args, **kwargs):
        """Add items to the queue.

        :param args: tuple is appended to list
        :param kwargs: are ignored.
        """
        self.queue.insert(0, args)



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Command line interfaces entrypoints."""


from flask_script import Manager

from . import account, check, githooks, prepare, repository


manager = Manager()

manager.add_command("account", account.manager)
manager.add_command("githooks", githooks.manager)
manager.add_command("prepare", prepare.manager)
manager.add_command("repository", repository.manager)
manager.add_command("check", check.manager)


def main():  # pragma: no cover
    """Running the manager."""
    from .. import create_app
    manager.app = create_app()
    manager.run()



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Tasks to run on the worker."""

from __future__ import unicode_literals

import os
import re
import shutil
import logging
import tempfile
import requests
import yaml
from flask import json

from .kwalitee import get_options, check_message, check_file
from .models import db, BranchStatus, CommitStatus
# Worker logger
LOGGER = logging.getLogger("rq.worker")


def get_headers(repository, config):
    """Get the HTTP headers for the GitHub api.

    This is required to post comments on GitHub on your behalf.
    Please update your configuration accordingly.

    .. code-block:: python

        ACCESS_TOKEN = "deadbeef..."

    It can also be overwritten per user.

    .. code-block:: console

        $ kwalitee account add username --token=deadbeef...

    :return: HTTP headers
    :rtype: dict

    """
    token = repository.owner.token or config["ACCESS_TOKEN"]
    return {
        "Content-Type": "application/json",
        "Authorization": "token {token}".format(token=token)
    }


def push(commit_status_id, commit_url, status_url, config):
    """Performing all the tests on the commit.

    :param commit_status_id: identifier of the branch status
    :param commit_url: github api commit url
    :param status_url: github api status url
    :param config: configuration dictionary

    """
    body = {}
    length = 0

    commit_status = CommitStatus.query.filter_by(id=commit_status_id).first()
    if not commit_status:
        raise AssertionError("Unknown commit: {0} ({1})"
                             .format(commit_status_id, commit_url))

    options = get_options(config)
    headers = get_headers(commit_status.repository, config)

    check_commit_messages = config.get("CHECK_COMMIT_MESSAGES", True)
    check_pep8 = options["pep8"]
    check_pep257 = options["pep257"]
    check_license = options["license"]
    check_files = check_pep8 or check_pep257 or check_license

    commit = requests.get(commit_url, headers=headers)
    data = json.loads(commit.content)

    # FIXME guessing the status URL based on the commit one.
    statuses_url = data["url"].replace("commits", "statuses")
    comments_url = data["comments_url"]
    sha = data["sha"]
    files = data["files"]

    # Check only if the component does not contain 'wip'.
    component = data["commit"]["message"].split(":", 1)[0]
    is_wip = bool(re.match(r"\bwip\b", component, re.IGNORECASE))
    check = config.get("CHECK_WIP", False) or not is_wip

    # This commit might have been checked in a pull request, which skipped
    # the files check.
    is_new = False
    checked = False
    if commit_status.is_pending() or commit_status.content["files"] is None:
        is_new = True

    if check and check_commit_messages:
        errs = _check_commit(commit_status, data, **options)
        length = len(errs)
        checked = True

    if check and is_new and checked and length:
        body = "\n".join(commit_status.content["message"])
        try:
            requests.post(comments_url,
                          data=json.dumps(dict(body=body.strip())),
                          headers=headers)
        except requests.RequestException:
            LOGGER.exception(comments_url)

    if check and is_new and check_files:
        tmp = tempfile.mkdtemp()
        filenames = _download_files_from_commit(files, sha, tmp)
        total, messages = _check_files(filenames, tmp, **options)
        shutil.rmtree(tmp)

        commit_status.content = dict(commit_status.content, files=messages)
        length += total

        if total:
            _post_file_comments(comments_url, messages, headers)

    if not check:
        commit_status.content = dict(commit_status.content, message=[])

    db.session.add(commit_status)
    db.session.commit()

    state = "error" if length > 0 else "success"
    body = dict(state=state,
                target_url=status_url,
                description="[{0}] {1} errors".format(state, length),
                context=config.get("CONTEXT"))

    # Do not send the status if we've already seen this commit in the past.
    if is_new:
        try:
            requests.post(statuses_url,
                          data=json.dumps(body),
                          headers=headers)
        except requests.RequestException:
            LOGGER.exception(statuses_url)
    return body


def pull_request(branch_status_id, pull_request_url, status_url, config):
    """Performing all the tests on the pull request.

    Then pings back the given status_url and update the issue labels.

    :param branch_status_id: identifier of the branch status.
    :type branch_status_id: int
    :param pull_request_url: github api pull request
    :type pull_request_url: str
    :param status_url: github api status url
    :type status_url: str
    :param config: configuration dictionary
    :type config: dict
    :return: status body and applied labels
    :rtype: dict

    """
    errors = {"message": [], "commits": [], "files": {}}
    length = 0

    branch_status = BranchStatus.query.filter_by(id=branch_status_id).first()

    if not branch_status:
        raise AssertionError("Unknown branch: {0} ({1})"
                             .format(branch_status_id, pull_request_url))
    if not branch_status.is_pending():
        LOGGER.info("Known pull request, skipping.")
        return branch_status.errors

    options = get_options(config)

    headers = get_headers(branch_status.commit.repository, config)

    pull_request = requests.get(pull_request_url, headers=headers)
    data = json.loads(pull_request.content)

    # Update config with .kwalitee.yml from git root folder
    options.update(_check_for_kwalitee_configuration(data))

    issue_url = data["issue_url"]
    commits_url = data["commits_url"]
    files_url = data["commits_url"].replace("/commits", "/files")
    review_comments_url = data["review_comments_url"]

    # Check only if the title does not contain 'wip'.
    is_wip = bool(re.match(r"\bwip\b", data["title"], re.IGNORECASE))
    check = config.get("CHECK_WIP", False) or not is_wip
    options["CHECK"] = check

    check_pep8 = options["pep8"]
    check_pep257 = options["pep257"]
    check_license = options["license"]
    check_files = check_pep8 or check_pep257 or check_license

    # Checking commits
    errs, messages = _check_commits(
        branch_status.commit.repository,
        commits_url,
        **options
    )

    errors["message"] += errs
    length += len(errs)

    for msg in messages:
        errors["commits"].append(msg["status"])
        body = "\n".join(msg["errors"]["message"])
        if body is not "":
            if db.session.is_modified(msg["status"]):
                db.session.add(msg["status"])
                try:
                    requests.post(msg["comments_url"],
                                  data=json.dumps(dict(body=body)),
                                  headers=headers)
                except requests.RequestException:
                    LOGGER.exception(msg["comments_url"])

    db.session.commit()

    # Checking the files
    if check and check_files:
        tmp = tempfile.mkdtemp()
        filenames = _download_files_from_pull_request(files_url, tmp)
        total, messages = _check_files(filenames, tmp, **options)
        shutil.rmtree(tmp)

        errors["files"].update(messages)
        length += total

        if total:
            _post_file_comments(review_comments_url, messages, headers)
        state = "error" if length > 0 else "success"
        body = dict(state=state,
                    target_url=status_url,
                    description="[{0}] {1} errors".format(state, length),
                    context=config.get("CONTEXT"))
        try:
            requests.post(data["statuses_url"],
                          data=json.dumps(body),
                          headers=headers)
        except requests.RequestException:
            LOGGER.exception(data["statuses_url"])

    branch_status.content = errors
    db.session.add(branch_status)
    db.session.commit()

    new_labels = set()
    if is_wip:
        new_labels = set([config.get("LABEL_WIP", "in_work")])
    if not new_labels:
        if length == 0:
            new_labels = set([config.get("LABEL_READY", "in_integration")])
        else:
            new_labels = set([config.get("LABEL_REVIEW", "in_review")])
    labels = _update_labels(issue_url, new_labels, headers, config)

    return dict(errors, labels=list(labels))


def _update_labels(issue_url, new_labels, headers, config):
    """Update the labels of an issue.

    :param issue_url: url of the issue to read the existing labels from.
    :param new_labels: set of new labels to be applied.
    :return: set of updated labels of the issue.
    """
    labels, labels_url = _get_issue_labels(issue_url)
    labels.discard(config.get("LABEL_WIP", "in_work"))
    labels.discard(config.get("LABEL_REVIEW", "in_review"))
    labels.discard(config.get("LABEL_READY", "in_integration"))

    labels.update(new_labels)
    try:
        requests.put(labels_url.replace("{/name}", ""),
                     data=json.dumps(list(labels)),
                     headers=headers)
    except requests.RequestException:
        LOGGER.exception(labels_url)

    return labels


def _post_file_comments(comments_url, messages, headers):
    """Post comment on each file."""
    # Sorting the filenames so the output is predictable and tests are easier.
    filenames = list(messages.keys())
    filenames.sort()
    for filename in filenames:
        msg = messages[filename]
        body = "\n".join(msg["errors"])
        if body is not "":
            # Comment on first line.
            position = 0
            try:
                requests.post(comments_url,
                              data=json.dumps(dict(body=body,
                                                   commit_id=msg["sha"],
                                                   path=filename,
                                                   position=position)),
                              headers=headers)
            except requests.RequestException:
                LOGGER.exception(comments_url)


def _download_files_from_pull_request(url, tmp):
    """Download the files of a pull request.

    :param url: pull request's list of files URL
    :param tmp: temporary directory to put the files into
    :return: filenames with their sha1
    """
    sha_match = re.compile(r"(?<=ref=)[^=]+")
    response = requests.get(url)
    files = json.loads(response.content)

    for f in files:
        filename = f["filename"]
        sha = sha_match.search(f["contents_url"]).group(0)

        _download_file(f["raw_url"], os.path.join(tmp, filename))

        yield (sha, filename)


def _download_files_from_commit(files, sha, tmp):
    """Download the files from the commit.

    :param files: list of files from the commit JSON
    :param tmp: temporary directory to put the files into
    :returns: filenames
    """
    for f in files:
        if f["status"] != "removed":
            filename = f["filename"]

            if _download_file(f["raw_url"], os.path.join(tmp, filename)):
                yield (sha, filename)


def _download_file(source, destination):
    """Download the given file and puts it at the desired place."""
    try:
        response = requests.get(source)

        # File not found or other error.
        if response.status_code > 400:
            return requests.HTTPError(response.status_code)

        dirname = os.path.dirname(destination)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        with open(destination, "wb+") as fp:
            for block in response.iter_content(1024):
                if block:
                    fp.write(block)

        return True
    except requests.RequestException:
        LOGGER.exception(source)
        return False


def _get_issue_labels(issue_url):
    """Download the labels of the issue."""
    issue = json.loads(requests.get(issue_url).content)
    labels = set()
    for label in issue["labels"]:
        labels.add(label["name"])
    return labels, issue["labels_url"]


def _check_commit(commit_status, commit, **kwargs):
    """Check one commit message.

    Does nothing if the commit has already been checked in the past.

    :param commit_status: commit status
    :type: CommitStatus
    :param commit: json object
    :type commit: dict
    :return: errors
    :rtype: list
    """
    errs = None

    if commit_status.is_pending():
        errs = check_message(commit["commit"]["message"], **kwargs)
        commit_status.content = dict(commit_status.content, message=errs)
    else:
        errs = commit_status.content["message"]

    return errs


def _check_commits(repository, url, **kwargs):
    """Check the commit messages of a pull request."""
    errors = []
    messages = []

    check = kwargs.get("CHECK")
    check_commit_messages = kwargs.get("CHECK_COMMIT_MESSAGES", True)

    response = requests.get(url)
    commits = json.loads(response.content)
    for commit in reversed(commits):
        errs = []
        sha = commit["sha"]
        url = commit["html_url"]

        commit_status = CommitStatus.query.filter_by(
            repository_id=repository.id,
            sha=sha).first()

        if not commit_status:
            raise AssertionError(
                "CommitStatus not found for {0.fullname} {sha}"
                .format(repo=repository, sha=sha))

        if check and check_commit_messages:
            errs = _check_commit(commit_status, commit, **kwargs)

        # filter out the needs more reviewers
        e = list(filter(lambda x: not re.match(r"\d+: M100\b", x), errs))

        messages.append({
            "sha": sha,
            "status": commit_status,
            "comments_url": commit["comments_url"],
            "errors": {
                "message": e,
                "files": {}
            }
        })
        errors += list(map(lambda x: "{0}: {1}".format(sha, x), errs))
    return errors, messages


def _check_files(files, cwd, **kwargs):
    """Download and runs the checks on the files of a pull request.

    Format of the dict returned.

    .. code-block:: json

        {
            "filename": {
                "sha": ...
                "errors": [...]
            },
            "filename2": { ... }
        }

    :param files: list of files ``(sha, filename)``.
    :param cwd: directory where the files are located.
    :param kwargs: arguments for :meth:`check_file`.
    :return: tuple composed of the number of errors found and a dict with the
             files checked.
    """
    total = 0
    messages = {}

    for sha, filename in files:
        errors = check_file(os.path.join(cwd, filename), **kwargs)

        messages[filename] = {
            "sha": sha,
            "errors": errors
        }
        total += len(errors or [])
    return total, messages


def _check_for_kwalitee_configuration(github_pr):
    """Check if a configuaration file exists in the repo.

    .. note::

        If the `.kwalitee.yml` file has not been found it will return
        an empty dict will not affect the process.

    :param dict github_pr: responded github pull_request object
    :return: yaml parsed configuration
    :rtype: dict
    """
    # download .kwalitee.yml if exists
    github_contents_api = (
        "https://api.github.com/repos/{name}/contents/.kwalitee.yml"
        "?ref={branch}"
    ).format(
        name=github_pr.get('base', {}).get('repo', {}).get('full_name'),
        branch=github_pr.get('base', {}).get('ref')
    )
    # try to request the file
    request_configuration = requests.get(github_contents_api)
    data = {}
    # check if the status code is 200
    if request_configuration.status_code == 200:
        response = request_configuration.content
        data_response = json.loads(response)
        # get the file's download url and download it
        download_url = data_response.get('download_url')
        # create a temporary file
        tmp = tempfile.mkdtemp()
        destination = os.path.join(tmp, data_response.get('name', 'temp'))
        # and finaly download the file into it
        written = _download_file(download_url, destination)
        # if there is an actual downloaded file try to read the yaml
        if written:
            with open(destination, 'r') as file_read:
                data = yaml.load(file_read.read())
        # finaly delete the tempdirectory
        shutil.rmtree(tmp)
    return data



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Configuration for py.test."""

from __future__ import unicode_literals

import os
import pytest
import shutil
import tempfile
import subprocess

from kwalitee import create_app
from kwalitee.models import db as _db, Account, Repository


@pytest.fixture(scope="session")
def app(request):
    """Session-wide test Flask application."""
    config = {
        "TESTING": True,
        "DATABASE_NAME": "testdb"
    }
    app = create_app("kwalitee", config)

    ctx = app.app_context()
    ctx.push()

    def teardown():
        ctx.pop()

    request.addfinalizer(teardown)
    return app


@pytest.fixture(scope="session")
def db(app, request):
    """Session-wide test database."""
    if os.path.exists(app.config["DATABASE"]):
        os.unlink(app.config["DATABASE"])

    def teardown():
        _db.drop_all()
        os.unlink(app.config["DATABASE"])

    _db.app = app
    _db.create_all()

    request.addfinalizer(teardown)
    return _db


@pytest.fixture(scope="function")
def session(db, request):
    """Create a new database session for a test."""
    connection = db.engine.connect()
    transaction = connection.begin()

    session = db.create_scoped_session()
    db.session = session

    def teardown():
        transaction.rollback()
        connection.close()
        session.remove()

    request.addfinalizer(teardown)
    return session


@pytest.fixture(scope="function")
def owner(session, request):
    """Create an account for a test."""
    owner = Account.find_or_create("invenio", token="DEADBEEF")

    def teardown():
        session.delete(owner)
        session.commit()

    request.addfinalizer(teardown)
    return owner


@pytest.fixture(scope="function")
def repository(owner, session, request):
    """Create a repository for a test."""
    repo = (Repository.find_or_create(owner, "test"))

    def teardown():
        session.delete(repo)
        session.commit()

    request.addfinalizer(teardown)
    return repo


@pytest.fixture(scope="function")
def git(request):
    """Create a git repository."""
    repo = tempfile.mkdtemp()

    commands = (("git", "init"),
                ("git", "config", "user.name", "Herp Derpson"),
                ("git", "config", "user.email", "herp.derpson@example.org"),
                ("touch", "README.rst"),
                ("git", "add", "README.rst"),
                ("git", "commit", "-m", "empty README"),
                ("git", "checkout", "-b", "testbranch"),
                ("touch", "TODO"),
                ("git", "add", "TODO"),
                ("git", "commit", "-m", "global: kikoo\n\nBy: bob <a@b.org>"),
                ("git", "checkout", "master"))

    for command in commands:
        status = subprocess.Popen(command,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  cwd=repo).wait()
        assert 0 == status, " ".join(command)

    def teardown():
        shutil.rmtree(repo)

    request.addfinalizer(teardown)
    return repo



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""
kwalitee base configuration.

To change it, put a config.py into your instance.

.. py:data:: CONTEXT

    Context for Github.

    .. seealso::

        `Github API Statuses
        <https://developer.github.com/v3/repos/statuses/>`_


.. py:data:: COMPONENTS

    List of supported components.

    .. seealso:: :func:`.kwalitee.check_message`

.. py:data:: ACCESS_TOKEN

    Github access token. Used to post statuses and comments. It MUST be set.

.. py:data:: AUTO_CREATE

    Allow anyone to add its repository by setting up the webhook on Github.

    **Default:** ``False``

.. py:data:: CHECK_COMMIT_MESSAGES

    Enable the :func:`commit message checks <.kwalitee.check_message>`.

    **Default:** ``True``

.. py:data:: CHECK_WIP

    Enable the *work-in-progress* pull requests checks. Disabled by default.

    **Default:** ``False``

.. py:data:: CHECK_LICENSE

    Enable the :func:`license checks <.kwalitee.check_license>`.

    **Default:** ``True``

.. py:data:: CHECK_PEP8

    Enable the :func:`PEP8 checks <.kwalitee.check_pep8>`.

    **Default:** ``True``

.. py:data:: CHECK_PEP257

    Enable the :func:`PEP257 checks <.kwalitee.check_pep257>`.

    **Default:** ``True``

.. py:data:: CHECK_PYFLAKES

    Enable the PyFlakes checks. PEP8 checks are required.

    **Default:** ``True``

.. py:data:: IGNORE

    Error codes to ignore.

    **Default:** ``['E123', 'E226', 'E24', 'E501', 'E265']``

.. py:data:: SELECT

    Error codes to specially enable.

    **Default:** ``[]``

.. py:data:: WORKER_TIMEOUT

    Background worker job time window.

    Any job taking longer than that will be killed.

    RQ default timeout is 180 seconds

.. py:data:: MIN_REVIEWERS

    Minimum number of reviewers for py:func:`message check
    <.kwalitee.check_message>`.

    **Default:** 3

.. py:data:: LABEL_WIP

    Label to apply for a *work-in-progress* pull request.

    **Default:** ``"in_work"``

.. py:data:: LABEL_REVIEW

    Label to apply for a pull request that needs more reviewers.

    **Default:** ``"in_review"``

.. py:data:: LABEL_READY

    Label to apply for a pull request that passes all the checks.

    **Default:** ``"in_integration"``

.. py:data:: EXCLUDES

    List of regex of excluded files.

    **Default:** ``[]``

"""

COMPONENTS = [
    'base',
    'celery',
    'global',
    'grunt',
    'installation',
    'utils',
    # modules
    'access',
    'accounts',
    'alerts',
    'apikeys',
    'authorlist',
    'authors',
    'baskets',
    'bulletin',
    'circulation',
    'classifier',
    'cloudconnector',
    'comments',
    'communities',
    'converter',
    'dashboard',
    'deposit',
    'documentation',
    'documents',
    'editor',
    'encoder',
    'export',
    'exporter',
    'formatter',
    'groups',
    'indexer',
    'jsonalchemy',
    'knowledge',
    'linkbacks',
    'matcher',
    'merger',
    'messages',
    'oaiharvester',
    'oairepository',
    'pidstore',
    'previewer',
    'previews',
    'ranker',
    'record_editor',
    'records',
    'redirector',
    'refextract',
    'scheduler',
    'search',
    'sequencegenerator',
    'sorter',
    'statistics',
    'submit',
    'sword',
    'tags',
    'textminer',
    'tickets',
    'upgrader',
    'uploader',
    'workflows',
]

CONTEXT = "kwalitee"


GITHUB = "https://github.com/"
"""Github base URL."""
GITHUB_REPO = GITHUB + "{account}/{repository}/"
"""Github repository URL template."""

# Background worker
# -----------------
# WORKER_TIMEOUT = 180

# Checks run on the files
# -----------------------
#
# Default values, uncomment to change:
# CHECK_COMMIT_MESSAGES = True
# CHECK_WIP = False
# CHECK_LICENSE = True
# CHECK_PEP8 = True
# CHECK_PEP257 = True
# CHECK_PYFLAKES = True # PyFlakes requires PEP8

# You may ignore some codes from PEP8, PEP257 and
# the license checks as well.
IGNORE = ['E123', 'E226', 'E24', 'E501', 'E265']
# SELECT = []

# Apply the tests only to the files matching those criteria.
PEP257_MATCH = "(?!test_).*\.py"
"""Files checked for PEP257 conformance."""
PEP257_MATCH_DIR = "[^\.].*"
"""Directories checkes for PEP257 conformance."""

# Minimal number of reviewers needed to accept a commit.
#
# Default value, uncomment to change:
# MIN_REVIEWERS = 3

TRUSTED_DEVELOPERS = []
"""Super developers who's code never fail."""

# List of recognized signatures
#
#
SIGNATURES = ('Signed-off-by', 'Co-authored-by', 'Tested-by', 'Reviewed-by',
              'Acked-by')
"""Authors and reviewers signatures."""
ALT_SIGNATURES = 'Reported-by',
"""Alternative signatures recognized but not counted as reviewers."""


# Labels applied to the pull request in case we are in the following states:
# - wip, the wip label has been found in the title
# - review, some commit need more reviewers
# - ready, none of the above
#
# Default values, uncomment to change:
# LABEL_WIP = "in_work"
# LABEL_REVIEW = "in_review"
# LABEL_READY = "in_integration"

# Hooks
# -----

COMMIT_MSG_TEMPLATE = """{component}: description (max 50 chars, using nouns)

* Detailed description formatted as a bullet list (using present tense).

Signed-off-by: {author}
{extra}"""
"""Template used to generate the commit message from the git hook."""


HOOK_TEMPLATE = """#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
from kwalitee import create_app
from kwalitee.hooks import {hook}

if __name__ == "__main__":
    with create_app().app_context():
        sys.exit({hook}(sys.argv))
"""
"""Template used to generate the git hooks, customize at will."""


COMMIT_MSG_LABELS = (
    ('SECURITY', 'Security fixes'),
    ('INCOMPATIBLE', 'Incompatible changes'),
    ('NEW', 'New features'),
    ('BETTER', 'Improved features'),
    ('FIX', 'Bug fixes'),
    ('NOTE', 'Notes'),
    ('AMENDS', None),
)



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#g
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Command-line tools for managing repositories."""

from __future__ import absolute_import, print_function, unicode_literals

import sys

from ..models import db, Account, Repository

from flask_script import Manager

manager = Manager(usage="repository management")


def _is_valid_repo(repository):
    """Say whether the repository is valid.

    It must be formed of the account name and repository name.
    """
    if repository.count("/") is not 1:
        print("{0} is not a valid repository "
              "(expected: owner/repository)".format(repository),
              file=sys.stderr)
        return False
    return True


@manager.command
def add(repository):
    """Add a repository to the list of authorized ones."""
    if not _is_valid_repo(repository):
        return 1
    account_name, repository_name = repository.split("/")

    acc = Account.find_or_create(account_name)
    Repository.find_or_create(acc, repository_name)
    print("{0} is now allowed to webhook kwalitee!".format(repository),
          file=sys.stderr)


@manager.command
def remove(repository):
    """Remove a repository from the list of authorized ones."""
    if not _is_valid_repo(repository):
        return 1
    account_name, repository_name = repository.split("/")

    acc = Account.query.filter_by(name=account_name).first()
    if not acc:
        return
    repo = Repository.query.filter_by(name=repository_name,
                                      owner_id=acc.id).first()
    if not repo:
        return

    db.session.delete(repo)
    db.session.commit()


@manager.command
def list():
    """List all the repositories."""
    for acc in Account.query.order_by(db.asc(Account.name)).all():
        for repo in acc.repositories:
            print(repo.fullname)



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Prepare release news from git log.

Prepares release news from git log messages, breaking release news
into (1) sections (e.g. Security fixes, detected from commit labels)
and (2) modules (e.g. search, detected from commit log headlines).
"""

from __future__ import absolute_import, print_function, unicode_literals

import itertools
import re
import sys
import textwrap
from collections import OrderedDict

from flask import current_app

from flask_script import Manager

from .check import _git_commits, _pygit2_commits

manager = Manager(usage=__doc__)


def analyse_body_paragraph(body_paragraph, labels=None):
    """Analyse commit body paragraph and return (label, message).

    >>> analyse_body_paragraph('* BETTER Foo and bar.',
    >>> ... {'BETTER': 'Improvements'})
    ('BETTER', 'Foo and bar.')
    >>> analyse_body_paragraph('* Foo and bar.')
    (None, 'Foo and bar.')
    >>> analyse_body_paragraph('Foo and bar.')
    (None, None)
    """
    # try to find leading label first:
    for label, dummy in labels:
        if body_paragraph.startswith('* ' + label):
            return (label, body_paragraph[len(label) + 3:].replace('\n  ',
                                                                   ' '))
    # no conformed leading label found; do we have leading asterisk?
    if body_paragraph.startswith('* '):
        return (None, body_paragraph[2:].replace('\n  ', ' '))
    # no leading asterisk found; ignore this paragraph silently:
    return (None, None)


def remove_ticket_directives(message):
    """Remove ticket directives like "(closes #123).

    >>> remove_ticket_directives('(closes #123)')
    '(#123)'
    >>> remove_ticket_directives('(foo #123)')
    '(foo #123)'
    """
    if message:
        message = re.sub(r'closes #', '#', message)
        message = re.sub(r'addresses #', '#', message)
        message = re.sub(r'references #', '#', message)
    return message


def amended_commits(commits):
    """Return those git commit sha1s that have been amended later."""
    # which SHA1 are declared as amended later?
    amended_sha1s = []
    for message in commits.values():
        amended_sha1s.extend(re.findall(r'AMENDS\s([0-f]+)', message))
    return amended_sha1s


def enrich_git_log_dict(messages, labels):
    """Enrich git log with related information on tickets."""
    for commit_sha1, message in messages.items():
        # detect module and ticket numbers for each commit:
        component = None
        title = message.split('\n')[0]
        try:
            component, title = title.split(":", 1)
            component = component.strip()
        except ValueError:
            pass  # noqa

        paragraphs = [analyse_body_paragraph(p, labels)
                      for p in message.split('\n\n')]
        yield {
            'sha1': commit_sha1,
            'component': component,
            'title': title.strip(),
            'tickets': re.findall(r'\s(#\d+)', message),
            'paragraphs': [
                (label, remove_ticket_directives(message))
                for label, message in paragraphs
            ],
        }


@manager.option('repository', default='.', nargs='?', help='repository path')
@manager.option('commit', metavar='<sha or branch>', nargs='?',
                default='HEAD', help='an integer for the accumulator')
@manager.option('-c', '--components', default=False, action="store_true",
                help='group components', dest='group_components')
def release(commit='HEAD', repository='.', group_components=False):
    """Generate release notes."""
    from ..kwalitee import get_options
    from ..hooks import _read_local_kwalitee_configuration
    options = get_options(current_app.config)
    options.update(_read_local_kwalitee_configuration(directory=repository))

    try:
        sha = 'oid'
        commits = _pygit2_commits(commit, repository)
    except ImportError:
        try:
            sha = 'hexsha'
            commits = _git_commits(commit, repository)
        except ImportError:
            print('To use this feature, please install pygit2. GitPython will '
                  'also work but is not recommended (python <= 2.7 only).',
                  file=sys.stderr)
            return 2

    messages = OrderedDict([(getattr(c, sha), c.message) for c in commits])

    for commit_sha1 in amended_commits(messages):
        if commit_sha1 in messages:
            messages.remove(commit_sha1)

    full_messages = list(
        enrich_git_log_dict(messages, options.get('commit_msg_labels'))
    )
    indent = '  ' if group_components else ''
    wrapper = textwrap.TextWrapper(
        width=70,
        initial_indent=indent + '- ',
        subsequent_indent=indent + '  ',
    )

    for label, section in options.get('commit_msg_labels'):
        if section is None:
            continue
        bullets = []
        for commit in full_messages:
            bullets += [
                {'text': bullet, 'component': commit['component']}
                for lbl, bullet in commit['paragraphs']
                if lbl == label and bullet is not None
            ]
        if len(bullets) > 0:
            print(section)
            print('-' * len(section))
            print()
            if group_components:
                for component, bullets in itertools.groupby(
                        bullets, lambda b: b['component']):
                    bullets = list(bullets)
                    if len(bullets) > 0:
                        print('+ ' + component)
                        print()
                    for bullet in bullets:
                        print(wrapper.fill(bullet['text']))
                    print()
            else:
                for bullet in bullets:
                    print(wrapper.fill(bullet['text']))
            print()

    return 0



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import unicode_literals

import pytest

from kwalitee.models import Account
from hamcrest import assert_that, equal_to, contains_string


@pytest.fixture(scope="function")
def accounts(session, request):
    names = "invenio", "test", "bob"

    a = []
    for name in names:
        a.append(Account.find_or_create(name))

    def teardown():
        for account in a:
            session.delete(account)
        session.commit()

    request.addfinalizer(teardown)
    return a


def test_simple_status(app, accounts):
    """GET / displays the accounts"""

    tester = app.test_client()
    response = tester.get("/")

    assert_that(response.status_code, equal_to(200))
    body = response.get_data(as_text=True)
    for account in accounts:
        assert_that(body, contains_string(account.name))



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Tests for the ping event."""

from __future__ import unicode_literals

from flask import json
from hamcrest import assert_that, equal_to


def test_ping(app):
    """POST /payload (ping) is ignored by kwalitee"""
    tester = app.test_client()
    response = tester.post("/payload", content_type="application/json",
                           headers=(("X-GitHub-Event", "ping"),
                                    ("X-GitHub-Delivery", "1")),
                           data=json.dumps({"hook_id": 1,
                                            "zen": "Responsive is better "
                                                   "than fast."}))
    assert_that(response.status_code, equal_to(200))


def test_ping_no_headers(app):
    """POST /payload (ping) expects a X-GitHub-Event header"""
    tester = app.test_client()
    response = tester.post("/payload",
                           data=json.dumps({"hook_id": 1,
                                            "zen": "Responsive is better "
                                                   "than fast."}))
    body = json.loads(response.data)
    assert_that(response.status_code, equal_to(500))
    assert_that(body["exception"],
                equal_to("No X-GitHub-Event HTTP header found"))
    assert_that(body["status"], equal_to("failure"))


def test_not_a_ping(app):
    """POST /payload (pong) rejects an unknown event"""
    tester = app.test_client()
    response = tester.post("/payload",
                           headers=(("X-GitHub-Event", "pong"),
                                    ("X-GitHub-Delivery", "1")),
                           data=json.dumps({"hook_id": 1,
                                            "zen": "Responsive is better "
                                                   "than fast."}))
    body = json.loads(response.data)
    assert_that(response.status_code, equal_to(500))
    assert_that(body["exception"],
                equal_to("Event pong is not supported"))
    assert_that(body["status"], equal_to("failure"))



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#g
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Command-line tools for the git hooks."""

from __future__ import absolute_import, print_function, unicode_literals

import os
import sys

from flask import current_app as app
from flask_script import Manager

from ..hooks import (pre_commit_hook, prepare_commit_msg_hook, run,
                     post_commit_hook)

HOOKS = {
    "pre-commit": pre_commit_hook.__name__,
    "prepare-commit-msg": prepare_commit_msg_hook.__name__,
    "post-commit": post_commit_hook.__name__,
}
HOOK_PATH = os.path.join(".git", "hooks")

manager = Manager(usage="install githooks for kwalitee checks")


@manager.option("-f", "--force",
                help="Overwrite existing hooks", default=False,
                action="store_true")
def install(force=False):
    """Install git hooks."""
    ret, git_dir, _ = run("git rev-parse --show-toplevel")
    if ret != 0:
        print("ERROR: Please run from within a GIT repository.",
              file=sys.stderr)
        return False
    git_dir = git_dir[0]

    hooks_dir = os.path.join(git_dir, HOOK_PATH)

    tpl = app.config["HOOK_TEMPLATE"]

    for hook, name in HOOKS.items():
        hook_path = os.path.join(hooks_dir, hook)
        if os.path.exists(hook_path) and not force:
            print("Hook already exists. Skipping {0}".format(hook_path),
                  file=sys.stderr)
            continue

        with open(hook_path, "w") as fh:
            fh.write(tpl.format(hook=name))
        os.chmod(hook_path, 0o755)
    return True


@manager.command
def uninstall():
    """Uninstall git hooks."""
    ret, git_dir, _ = run("git rev-parse --show-toplevel")
    if ret != 0:
        print("ERROR: Please run from within a GIT repository.",
              file=sys.stderr)
        return False
    git_dir = git_dir[0]

    hooks_dir = os.path.join(git_dir, HOOK_PATH)

    for hook, name in HOOKS.items():
        hook_path = os.path.join(hooks_dir, hook)
        if os.path.exists(hook_path):
            os.remove(hook_path)
    return True



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""kwalitee Flask application and git hooks."""

from __future__ import unicode_literals

from .factory import create_app
from .version import __version__


__docformat__ = "restructuredtext en"
__all__ = ("__version__", "create_app")



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Initialize Redis and setups the RQ worker."""

from __future__ import absolute_import

from os import environ
from redis import Redis
from rq import Connection, Queue, Worker

conn = Redis(host=environ.get('REDIS_HOST', 'localhost'))


def init_app(app):
    """Initialize the RQ queue."""
    queue = Queue(connection=conn)
    app.config["queue"] = queue
    return app


if __name__ == "__main__":
    import sys
    import logging

    from .wsgi import application

    with application.app_context():
        if tuple(sys.version_info) < (2, 7):
            logger = logging.getLogger("rq.worker")
            logger.setLevel(logging.DEBUG)
            logger.addHandler(logging.StreamHandler())
        with Connection(conn):
            worker = Worker(list(map(Queue, ('high', 'default', 'low'))))
            worker.work()



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import unicode_literals

import sys
import pytest
from hamcrest import assert_that, equal_to, has_item, has_items

from kwalitee.cli.check import message

try:
    import pygit2  # noqa
    pygit = True
except ImportError:
    pygit = False

skip = pytest.mark.skipif(sys.version_info > (3, 0) and not pygit,
                          reason="no pygit2 and not GitPython")


@skip
def test_check_head(capsys, session, app, git):
    app.config['COLORS'] = False

    assert_that(message("HEAD", repository=git), equal_to(1))

    out, _ = capsys.readouterr()
    assert_that(out.split("\n"),
                has_items("1: M110 missing component name",
                          "1: M101 signature is missing",
                          "1: M100 needs more reviewers"))


@skip
def test_check_branch(capsys, session, app, git):
    app.config['TRUSTED_DEVELOPERS'] = ('a@b.org',)
    app.config['SIGNATURES'] = ('By',)
    app.config['COMPONENTS'] = ('global',)
    app.config['COLORS'] = False

    assert_that(message("master..testbranch", repository=git), equal_to(0))

    out, _ = capsys.readouterr()
    assert_that(out.split("\n"), has_item("Everything is OK."))


@skip
def test_check_branch_wrong_side(capsys, session, app, git):
    app.config['COLORS'] = False

    assert_that(message("testbranch..master", repository=git), equal_to(0))

    out, _ = capsys.readouterr()
    assert_that(out.strip(), equal_to(""))



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Git hooks."""

from __future__ import print_function, unicode_literals

import os
import re
import shutil
import sys

from codecs import open
from subprocess import PIPE, Popen
from tempfile import mkdtemp

import yaml

from .kwalitee import SUPPORTED_FILES, check_file, check_message, get_options


def _get_files_modified():
    """Get the list of modified files that are Python or Jinja2."""
    cmd = "git diff-index --cached --name-only --diff-filter=ACMRTUXB HEAD"
    _, files_modified, _ = run(cmd)

    extensions = [re.escape(ext) for ext in list(SUPPORTED_FILES) + [".rst"]]
    test = "(?:{0})$".format("|".join(extensions))
    return list(filter(lambda f: re.search(test, f), files_modified))


def _get_git_author():
    """Return the git author from the git variables."""
    _, stdout, _ = run("git var GIT_AUTHOR_IDENT")

    git_author = stdout[0]
    return git_author[:git_author.find(">") + 1]


def _get_component(filename, default="global"):
    """Get component name from filename."""
    if hasattr(filename, "decode"):
        filename = filename.decode()
    parts = filename.split(os.path.sep)

    if len(parts) >= 3:
        if parts[1] in "modules legacy ext".split():
            return parts[2]
    if len(parts) >= 2:
        if parts[1] in "base celery utils".split():
            return parts[1]
    if len(parts) >= 1:
        if parts[0] in "grunt docs".split():
            return parts[0]
    return default


def _get_components(files):
    """Compile the components list from the given files."""
    return tuple(set(_get_component(f) for f in files))


def _prepare_commit_msg(tmp_file, author, files_modified=None, template=None):
    """Prepare the commit message in tmp_file.

    It will build the commit message prefilling the component line, as well
    as the signature using the git author and the modified files.

    The file remains untouched if it is not empty.
    """
    files_modified = files_modified or []
    template = template or "{component}:\n\nSigned-off-by: {author}\n{extra}"
    if hasattr(template, "decode"):
        template = template.decode()

    with open(tmp_file, "r", "utf-8") as fh:
        contents = fh.readlines()
        msg = filter(lambda x: not (x.startswith("#") or x.isspace()),
                     contents)
        if len(list(msg)):
            return

    component = "unknown"
    components = _get_components(files_modified)

    if len(components) == 1:
        component = components[0]
    elif len(components) > 1:
        component = "/".join(components)
        contents.append(
            "# WARNING: Multiple components detected - consider splitting "
            "commit.\r\n"
        )

    with open(tmp_file, "w", "utf-8") as fh:
        fh.write(template.format(component=component,
                                 author=author,
                                 extra="".join(contents)))


def _check_message(message, options):
    """Checking the message and printing the errors."""
    options = options or dict()

    from flask import current_app
    with current_app.app_context():
        options.update(get_options(current_app.config))

    errors = check_message(message, **options)

    if errors:
        for error in errors:
            print(error, file=sys.stderr)

        return False
    return True


def prepare_commit_msg_hook(argv):
    """Hook: prepare a commit message."""
    from flask import current_app
    with current_app.app_context():
        options = get_options(current_app.config)

    # Check if the repo has a configuration repo
    options.update(_read_local_kwalitee_configuration())

    _prepare_commit_msg(argv[1],
                        _get_git_author(),
                        _get_files_modified(),
                        options.get('template'))
    return 0


def commit_msg_hook(argv):
    """Hook: for checking commit message (prevent commit)."""
    with open(argv[1], "r", "utf-8") as fh:
        message = "\n".join(filter(lambda x: not x.startswith("#"),
                                   fh.readlines()))
    options = {"allow_empty": True}

    if not _check_message(message, options):
        print("Aborting commit due to commit message errors (override with "
              "'git commit --no-verify').", file=sys.stderr)
        return 1
    return 0


def post_commit_hook(argv=None):
    """Hook: for checking commit message."""
    _, stdout, _ = run("git log -1 --format=%B HEAD")
    message = "\n".join(stdout)
    options = {"allow_empty": True}

    if not _check_message(message, options):
        print("Commit message errors (fix with 'git commit --amend').",
              file=sys.stderr)
        return 1
    return 0


def _read_local_kwalitee_configuration(directory="."):
    """Check if the repo has a ``.kwalitee.yaml`` file."""
    filepath = os.path.abspath(os.path.join(directory, '.kwalitee.yml'))
    data = {}
    if os.path.exists(filepath):
        with open(filepath, 'r') as file_read:
            data = yaml.load(file_read.read())
    return data

# =============================================================================
# _pre_commit, pre_commit_hook() and run() is based on initially on Flake8
# git_hook, which is covered by the license:

# Copyright (C) 2014, 2015 CERN
# Copyright (C) 2011-2013 Tarek Ziade <tarek@ziade.org>
# Copyright (C) 2012-2013 Ian Cordasco <graffatcolmingov@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


def _pre_commit(files, options):
    """Run the check on files of the added version.

    They might be different than the one on disk. Equivalent than doing a git
    stash, check, and git stash pop.
    """
    errors = []
    tmpdir = mkdtemp()
    files_to_check = []
    try:
        for (file_, content) in files:
            # write staged version of file to temporary directory
            dirname, filename = os.path.split(os.path.abspath(file_))
            prefix = os.path.commonprefix([dirname, tmpdir])
            dirname = os.path.relpath(dirname, start=prefix)
            dirname = os.path.join(tmpdir, dirname)
            if not os.path.isdir(dirname):
                os.makedirs(dirname)
            filename = os.path.join(dirname, filename)
            with open(filename, "wb") as fh:
                fh.write(content)
            files_to_check.append((file_, filename))

        for (file_, filename) in files_to_check:
            errors += list(map(lambda x: "{0}: {1}".format(file_, x),
                               check_file(filename, **options) or []))
    finally:
        shutil.rmtree(tmpdir, ignore_errors=True)

    return errors


def pre_commit_hook(argv=None):
    """Hook: checking the staged files."""
    from flask import current_app
    with current_app.app_context():
        options = get_options(current_app.config)

    # Check if the repo has a configuration repo
    options.update(_read_local_kwalitee_configuration())

    files = []
    for filename in _get_files_modified():
        # get the staged version of the file and
        # write the staged version to temp dir with its full path to
        # avoid overwriting files with the same name
        _, stdout, _ = run("git show :{0}".format(filename), raw_output=True)
        files.append((filename, stdout))

    errors = _pre_commit(files, options)

    for error in errors:
        if hasattr(error, "decode"):
            error = error.decode()
        print(error, file=sys.stderr)

    if errors:
        print("Aborting commit due to kwalitee errors (override with "
              "'git commit --no-verify').",
              file=sys.stderr)
        return 1
    return 0


def run(command, raw_output=False):
    """Run a command using subprocess.

    :param command: command line to be run
    :type command: str
    :param raw_output: does not attempt to convert the output as unicode
    :type raw_output: bool
    :return: error code, output (``stdout``) and error (``stderr``)
    :rtype: tuple

    """
    p = Popen(command.split(), stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    # On python 3, subprocess.Popen returns bytes objects.
    if not raw_output:
        return (
            p.returncode,
            [line.rstrip() for line in stdout.decode("utf-8").splitlines()],
            [line.rstrip() for line in stderr.decode("utf-8").splitlines()]
        )
    else:
        return (p.returncode, stdout, stderr)
# =============================================================================



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import unicode_literals

import os
import tempfile

from hamcrest import (assert_that, has_length, has_item, has_items, is_not,
                      contains_string)
from kwalitee.kwalitee import check_pep8, check_pep257, check_license
from unittest import TestCase


class TestCheckFile(TestCase):

    def setUp(self):
        fixtures = os.path.join(os.path.dirname(__file__), "fixtures", "")
        self.valid = "{0}valid.py.test".format(fixtures)
        self.invalid = "{0}invalid.py.test".format(fixtures)
        self.invalid_whitespaces = "{0}invalid_whitespaces.py.test".format(
            fixtures)
        self.invalid_token = "{0}invalid_token.py.test".format(fixtures)
        self.invalid_all = "{0}invalid_all.py.test".format(fixtures)
        self.error = "{0}error.py.test".format(fixtures)
        self.empty = "{0}empty.py.test".format(fixtures)
        self.invalid_license = "{0}invalid_license.py.test".format(fixtures)
        self.valid_license = "{0}valid_license.py.test".format(fixtures)
        self.missing_license = "{0}missing_license.py.test".format(fixtures)
        self.license_html = "{0}license.html.test".format(fixtures)
        self.license_js = "{0}license.js.test".format(fixtures)
        self.license_css = "{0}license.css.test".format(fixtures)
        self.cp1252 = "{0}cp1252.py.test".format(fixtures)


class TestCheckPep8(TestCheckFile):

    """Unit tests of the PEP8 check."""

    def test_valid(self):
        """valid.py is correctly formatted (according to pep8)"""
        errors = check_pep8(self.valid, ignore=('E265', 'E266'))
        assert_that(errors, has_length(0))

    def test_invalid_file(self):
        """invalid.py has 7 PEP8 violations + 1 from pyFlakes"""
        errors = check_pep8(self.invalid)
        assert_that(errors, has_length(8))

    def test_erroneous_file(self):
        """error.py has 2 pyflakes violations"""
        errors = check_pep8(self.error)
        assert_that(errors, has_length(2))

    def test_ignore(self):
        """ignored PEP8 codes are ignored"""
        errors = check_pep8(self.invalid,
                            ignore=('E111', 'E113', 'E901'))
        assert_that(errors, has_length(0))

    def test_ignore_license(self):
        """ignored PEP8 codes are ignored"""
        errors = check_pep8(self.error, ignore=('E265',))
        assert_that(errors, has_length(2))

    def test_select(self):
        """selected PEP8 codes are selected"""
        errors = check_pep8(self.invalid, select=('E111',))
        assert_that(errors, has_length(3))


class TestCheckPep257(TestCheckFile):
    """Unit tests of the PEP257 checks."""

    def test_valid(self):
        """valid.py is correctly formatted (according to pep257)"""
        errors = check_pep257(self.valid)
        assert_that(errors, has_length(0))

    def test_invalid_whitespaces(self):
        """invalid_whitespaces.py is incorrectly formatted."""
        errors = check_pep257(self.invalid_whitespaces)
        assert_that(errors, has_length(5))

    def test_missing(self):
        """invalid.py has no docstring"""
        errors = check_pep257(self.invalid)
        assert_that(errors, has_item("1: D100 Missing docstring in public module"))

    def test_invalid_token(self):
        """invalid_token.py has a tokenization error"""
        errors = check_pep257(self.invalid_token)
        assert_that(errors, has_item("6:0 EOF in multi-line string"))

    def test_invalid_all(self):
        """invalid_all.py has a mutable all"""
        errors = check_pep257(self.invalid_all)
        assert_that(errors, has_length(1))
        assert_that(errors[0], contains_string("Could not evaluate contents of"
                                               " __all__. That means pep257"))

    def test_ignore(self):
        """ignored PEP257 codes are ignored"""
        errors = check_pep257(self.invalid, ignore=('D100'))
        assert_that(errors, has_length(0))

    def test_match(self):
        """test only the file that are matched by the regex"""
        errors = check_pep257("test_bar.py", match="(?!test_).*\.py")
        assert_that(errors, has_length(0))

    def test_match_dir(self):
        """test only the directories that are matched by the regex"""
        errors = check_pep257("foo/.hidden/spam/eggs/bar.py",
                              match_dir="[^\.].*")
        assert_that(errors, has_length(0))

    def test_match_absolute_dir(self):
        fp, path = tempfile.mkstemp(text=True)
        os.write(fp, "# -*- coding: utf-8 -*-\n".encode("ascii"))
        os.close(fp)

        errors = check_pep257(path, match_dir="[^\.].*")
        assert_that(errors, has_length(1))

        os.unlink(path)


class TestCheckLicense(TestCheckFile):

    """Unit tests of the license validation."""

    def test_license(self):
        """valid_license has a well formatted license."""
        errors = check_license(self.valid_license, year=2014)
        assert_that(errors, has_length(0))

    def test_missing(self):
        """missing_license has only the copyright"""
        errors = check_license(self.missing_license)
        assert_that(errors, has_item("13: L100 license is missing"))

    def test_license_jinja(self):
        """license.html has a well formatted license."""
        errors = check_license(self.license_html, year=2014)
        assert_that(errors, has_length(0))

    def test_license_js(self):
        """license.js has a well formatted license."""
        errors = check_license(self.license_js, year=2014, python_style=False)
        assert_that(errors, has_length(0))

    def test_license_css(self):
        """license.css has a well formatted license."""
        errors = check_license(self.license_css, year=2014, python_style=False)
        assert_that(errors, has_length(0))

    def test_empty(self):
        """empty files like __init__ are skipped"""
        errors = check_license(self.empty)
        assert_that(errors, has_length(0))

    def test_unicode_license(self):
        """invalid_license uses unicode copyright and multiline years.

        Where copyright is ©.
        """
        errors = check_license(self.invalid_license, year=2014)
        assert_that(errors, is_not(has_item("25: L101 copyright is missing")))

    def test_no_license(self):
        """license is missing"""
        errors = check_license(self.valid)
        assert_that(errors, has_item("25: L101 copyright is missing"))

    def test_no_copyright(self):
        """copyright is missing"""
        errors = check_license(self.valid)
        assert_that(errors, has_item("25: L101 copyright is missing"))

    def test_outdated_license(self):
        """valid_license has an outdated license."""
        errors = check_license(self.valid_license, year=2015)
        assert_that(errors, has_item("5: L102 copyright year is outdated, "
                                     "expected 2015 but got 2014"))

    def test_doesnt_look_like_gnu_gpl(self):
        """invalid_license doesn't look like the GNU GPL"""
        errors = check_license(self.invalid_license, year=2014)
        assert_that(errors, has_item("25: L103 license is not GNU GPLv2"))

    def test_badly_encoded_file(self):
        errors = check_license(self.cp1252, year=2014)
        assert_that(errors,
                    has_item("24: L190 file cannot be decoded as utf-8"))



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Configuration for py.test in views."""

import pytest

from kwalitee.models import Repository, CommitStatus, BranchStatus


@pytest.fixture(scope="function")
def repositories(owner, session, request):
    """Create a couple of repositories for a test."""
    repos = []
    names = "invenio", "test", "bob"
    for name in names:
        repos.append(Repository.find_or_create(owner, name))

    def teardown():
        for repo in repos:
            session.delete(repo)
        session.commit()

    request.addfinalizer(teardown)
    return repos


@pytest.fixture(scope="function")
def commits(owner, repository, session, request):
    """Create a couple of commits for a test."""
    cs = [
        {"sha": "ef01234"},
        {"sha": "789abcd"},
        {"sha": "0123456"},
    ]
    url_template = "https://github.com/commits/{sha}"

    commits = []
    for commit in cs:
        cs = CommitStatus(repository,
                          commit["sha"],
                          url_template.format(**commit),
                          {"message": [], "files": None})
        commits.append(cs)
        session.add(cs)
    session.commit()

    def teardown():
        for commit in commits:
            session.delete(commit)
        session.commit()

    request.addfinalizer(teardown)
    return commits


@pytest.fixture(scope="function")
def branch(owner, repository, commits, session, request):
    """Create a branch status for a test."""
    branch_name = "test:wip/my-branch"

    bs = BranchStatus(commits[-1],
                      branch_name,
                      "https://github.com/pulls/1",
                      {"commits": commits, "files": {}})
    session.add(bs)
    session.commit()

    def teardown():
        session.delete(bs)
        session.commit()

    request.addfinalizer(teardown)
    return bs



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

from __future__ import unicode_literals

import os
import sys
import pytest
import shutil
import tempfile
import subprocess
from io import StringIO
from unittest import TestCase
from hamcrest import assert_that, has_length, is_not
from kwalitee.cli.githooks import install, uninstall, HOOK_PATH


@pytest.mark.usefixtures("session")
class GithookCliTest(TestCase):

    def setUp(self):
        self.hooks = ('pre-commit', 'prepare-commit-msg', 'post-commit')
        self.path = tempfile.mkdtemp()
        self.cwd = os.getcwd()
        self.stderr = sys.stderr
        sys.stderr = StringIO("")
        os.chdir(self.path)

    def tearDown(self):
        os.chdir(self.cwd)
        sys.stderr = self.stderr
        shutil.rmtree(self.path)

    def call(self, *args):
        return subprocess.Popen(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=self.path
        ).wait()

    def test_install_hooks(self):
        precommit = os.path.join(HOOK_PATH, "pre-commit")
        self.call("git", "init")
        self.call("touch", precommit)

        assert_that(install())

        for hook in self.hooks:
            filename = os.path.join(self.path, HOOK_PATH, hook)
            assert_that(os.path.exists(filename), filename)

        with open(precommit, "r") as f:
            assert_that(f.read(), has_length(0),
                        "precommit should be have been left as is")

        assert_that(uninstall())

        for hook in self.hooks:
            filename = os.path.join(self.path, HOOK_PATH, hook)
            assert_that(not os.path.exists(filename), filename)

    def test_dont_install_in_non_git(self):
        assert_that(not install())

        for hook in self.hooks:
            filename = os.path.join(self.path, HOOK_PATH, hook)
            assert_that(not os.path.exists(filename), filename)

        assert_that(not uninstall())

    def test_install_hooks_and_overrides(self):
        precommit = os.path.join(HOOK_PATH, "pre-commit")
        self.call("git", "init")
        self.call("touch", precommit)

        assert_that(install(force=True))

        for hook in self.hooks:
            filename = os.path.join(self.path, HOOK_PATH, hook)
            assert_that(os.path.exists(filename), filename)

        with open(precommit, "r") as f:
            assert_that(f.read(), is_not(has_length(0)),
                        "precommit should have been overridden")

        assert_that(uninstall())

        for hook in self.hooks:
            filename = os.path.join(self.path, HOOK_PATH, hook)
            assert_that(not os.path.exists(filename), filename)



# -*- coding: utf-8 -*-
#
# This file is part of kwalitee
# Copyright (C) 2014, 2015 CERN.
#
# kwalitee is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# kwalitee is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kwalitee; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Command-line tools for checking commits."""

from __future__ import absolute_import, print_function, unicode_literals

import colorama
import os
import re
import shutil
import sys

from flask import current_app
from flask_script import Manager

from tempfile import mkdtemp

manager = Manager(usage='check commits')


def _git_commits(commit, repository):
    import git
    cwd = os.getcwd()
    os.chdir(repository)
    g = git.Repo('.')
    kwargs = {'with_keep_cwd': True}
    if '..' not in commit:
        kwargs['max_count'] = 1
    commits = list(g.iter_commits(commit, **kwargs))

    os.chdir(cwd)
    return commits


def _pygit2_commits(commit, repository):
    from pygit2 import Repository, GIT_SORT_TOPOLOGICAL
    g = Repository(repository)

    if '..' in commit:
        tail, head = commit.split('..', 2)
        head = head or 'HEAD'
    else:
        head = commit
        tail = commit + '^'

    walker = g.walk(g.revparse_single(head).oid, GIT_SORT_TOPOLOGICAL)

    try:
        walker.hide(g.revparse_single(tail).oid)
    except KeyError:
        pass

    return walker


@manager.option('repository', default='.', nargs='?', help='repository path')
@manager.option('commit', metavar='<sha or branch>', nargs='?',
                default='HEAD', help='an integer for the accumulator')
def message(commit='HEAD', repository='.'):
    """Check the messages of the commits."""
    from ..kwalitee import check_message, get_options
    from ..hooks import _read_local_kwalitee_configuration
    options = get_options(current_app.config)
    options.update(_read_local_kwalitee_configuration(directory=repository))

    if options.get('colors') is not False:
        colorama.init(autoreset=True)
        reset = colorama.Style.RESET_ALL
        yellow = colorama.Fore.YELLOW
        green = colorama.Fore.GREEN
        red = colorama.Fore.RED
    else:
        reset = yellow = green = red = ''

    try:
        sha = 'oid'
        commits = _pygit2_commits(commit, repository)
    except ImportError:
        try:
            sha = 'hexsha'
            commits = _git_commits(commit, repository)
        except ImportError:
            print('To use this feature, please install pygit2. GitPython will '
                  'also work but is not recommended (python <= 2.7 only).',
                  file=sys.stderr)
            return 2

    template = '{0}commit {{commit.{1}}}{2}\n\n'.format(yellow, sha, reset)
    template += '{message}{errors}'

    count = 0
    ident = '    '
    re_line = re.compile('^', re.MULTILINE)
    for commit in commits:
        message = commit.message
        errors = check_message(message, **options)
        message = re.sub(re_line, ident, message)
        if errors:
            count += 1
            errors.insert(0, red)
        else:
            errors = [green, 'Everything is OK.']
        errors.append(reset)

        print(template.format(commit=commit,
                              message=message,
                              errors='\n'.join(errors)))

    return min(count, 1)


@manager.option('repository', default='.', nargs='?', help='repository path')
@manager.option('commit', metavar='<sha or branch>', nargs='?',
                default='HEAD', help='an integer for the accumulator')
def files(commit='HEAD', repository='.'):
    """Check the files of the commits."""
    from ..kwalitee import check_file, get_options, SUPPORTED_FILES
    from ..hooks import run, _read_local_kwalitee_configuration
    options = get_options(current_app.config)
    options.update(_read_local_kwalitee_configuration(directory=repository))

    if options.get('colors') is not False:
        colorama.init(autoreset=True)
        reset = colorama.Style.RESET_ALL
        yellow = colorama.Fore.YELLOW
        green = colorama.Fore.GREEN
        red = colorama.Fore.RED
    else:
        reset = yellow = green = red = ''

    try:
        sha = 'oid'
        commits = _pygit2_commits(commit, repository)
    except ImportError:
        try:
            sha = 'hexsha'
            commits = _git_commits(commit, repository)
        except ImportError:
            print('To use this feature, please install pygit2. GitPython will '
                  'also work but is not recommended (python <= 2.7 only).',
                  file=sys.stderr)
            return 2

    template = '{0}commit {{commit.{1}}}{2}\n\n'.format(yellow, sha, reset)
    template += '{message}{errors}\n'

    error_template = '\n{0}{{filename}}\n{1}{{errors}}{0}'.format(reset, red)
    no_errors = ['\n{0}Everything is OK.{1}'.format(green, reset)]
    msg_file_excluded = '\n{0}{{filename}} excluded.{1}'.format(yellow, reset)

    def _get_files_modified(commit):
        """Get the list of modified files that are Python or Jinja2."""
        cmd = "git show --no-commit-id --name-only --diff-filter=ACMRTUXB {0}"
        _, files_modified, _ = run(cmd.format(commit))

        extensions = [re.escape(ext)
                      for ext in list(SUPPORTED_FILES) + [".rst"]]
        test = "(?:{0})$".format("|".join(extensions))
        return list(filter(lambda f: re.search(test, f), files_modified))

    def _ensure_directory(filename):
        dir_ = os.path.dirname(filename)
        if not os.path.exists(dir_):
            os.makedirs(dir_)

    def _format_errors(args):
        filename, errors = args
        if errors is None:
            return msg_file_excluded.format(filename=filename)
        else:
            return error_template.format(filename=filename, errors='\n'.join(
                errors if len(errors) else no_errors))

    count = 0
    ident = '    '
    re_line = re.compile('^', re.MULTILINE)
    for commit in commits:
        message = commit.message
        commit_sha = getattr(commit, sha)
        tmpdir = mkdtemp()
        errors = {}
        try:
            for filename in _get_files_modified(commit):
                cmd = "git show {commit_sha}:{filename}"
                _, out, _ = run(cmd.format(commit_sha=commit_sha,
                                           filename=filename),
                                raw_output=True)

                destination = os.path.join(tmpdir, filename)
                _ensure_directory(destination)

                with open(destination, 'w+') as f:
                    f.write(out)

                errors[filename] = check_file(destination, **options)
        finally:
            shutil.rmtree(tmpdir, ignore_errors=True)

        message = re.sub(re_line, ident, message)
        if len(errors):
            count += 1
            errors = map(_format_errors, errors.items())
        else:
            errors = no_errors

        print(template.format(commit=commit,
                              message=message,
                              errors='\n'.join(errors)))

    return min(count, 1)
